"""This defines the Value object used by components and events."""

from .events import Event


class Value:
    """
    Create a new future Value Object

    Creates a new future Value Object which is used by Event Objects and the
    Manager to store the result(s) of an Event Handler's exeuction of some
    Event in the system.

    :param event: The Event this Value is associated with.
    :type  event: Event instance

    :param manager: The Manager/Component used to trigger notifications.
    :type  manager: A Manager/Component instance.

    :ivar result: True if this value has been changed.
    :ivar errors: True if while setting this value an exception occurred.
    :ivar notify: True or an event name to notify of changes to this value

    This is a Future/Promise implementation.
    """

    def __init__(self, event=None, manager=None):
        self.event = event
        self.manager = manager

        self.notify = False
        self.promise = False

        self.result = False
        self.errors = False
        self.parent = self
        self.handled = False

        self._value = None

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['manager']
        return odict

    def __contains__(self, y):
        value = self.value
        return y in value if isinstance(value, list) else y == value

    def __getitem__(self, y):
        v = self.value[y]
        if isinstance(v, Value):
            return v.value
        return v

    def __iter__(self):
        return iter((v.value if isinstance(v, Value) else v for v in self.value))

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        value = ''
        if self.result:
            value = repr(self.value)

        format = '<Value (%s) result=%r; errors=%r; for %r>'
        return format % (value, self.result, self.errors, self.event)

    def __str__(self):
        """x.__str__() <==> str(x)"""
        return str(self.value)

    def inform(self, force=False):
        if self.promise and not force:
            return

        notify = getattr(self.event, 'notify', False) or self.notify

        if self.manager is not None and notify:
            e = Event.create(notify, self) if isinstance(notify, str) else self.event.child('value_changed', self)

            self.manager.fire(e, self.manager)

    def getValue(self, recursive=True):
        value = self._value

        if not recursive:
            return value

        while isinstance(value, Value):
            value = value._value

        return value

    def setValue(self, value):
        if isinstance(value, Value):
            value.parent = self

        if self.result and isinstance(self._value, list):
            self._value.append(value)
        elif self.result:
            self._value = [self._value]
            self._value.append(value)
        else:
            self._value = value

        def update(o, v):
            if isinstance(v, Value):
                o.errors = v.errors
                o.result = v.result
            elif v is not None:
                o.result = True

                o.inform()

            if o.parent is not o:
                o.parent.errors = o.errors
                o.parent.result = o.result
                update(o.parent, v)

        update(self, value)

    value = property(getValue, setValue, None, 'Value of this Value')



#!/usr/bin/env python
from circuits import Component, Event


class hello(Event):
    """hello Event"""


class foo(Event):
    """foo Event"""


class bar(Event):
    """bar Event"""


class App(Component):
    def foo(self):
        return 1

    def bar(self):
        return 2

    def hello(self):
        x = yield self.call(foo())
        y = yield self.call(bar())
        yield x.value + y.value

    def started(self, component):
        x = yield self.call(hello())
        print(f'{x.value:d}')
        self.stop()


App().run()



#!/usr/bin/env python
from circuits.web import Controller
from circuits.web.wsgi import Application

from .helpers import urlopen


class Root(Controller):
    def index(self):
        yield 'Hello '
        yield 'World!'


application = Application() + Root()


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'
    assert f.headers.get('Transfer-Encoding') != 'chunked'



"""
Web Servers

This module implements the several Web Server components.
"""

from sys import stderr

from circuits import io
from circuits.core import BaseComponent, Timer, handler
from circuits.net.events import close, read, write
from circuits.net.sockets import BUFSIZE, TCPServer, UNIXServer

from .dispatchers import Dispatcher
from .events import terminate
from .http import HTTP


class BaseServer(BaseComponent):
    """
    Create a Base Web Server

    Create a Base Web Server (HTTP) bound to the IP Address / Port or
    UNIX Socket specified by the 'bind' parameter.

    :ivar server: Reference to underlying Server Component

    :param bind: IP Address / Port or UNIX Socket to bind to.
    :type bind: Instance of int, list, tuple or str

    The 'bind' parameter is quite flexible with what valid values it accepts.

    If an int is passed, a TCPServer will be created. The Server will be
    bound to the Port given by the 'bind' argument and the bound interface
    will default (normally to  "0.0.0.0").

    If a list or tuple is passed, a TCPServer will be created. The
    Server will be bound to the Port given by the 2nd item in the 'bind'
    argument and the bound interface will be the 1st item.

    If a str is passed and it contains the ':' character, this is
    assumed to be a request to bind to an IP Address / Port. A TCpServer
    will thus be created and the IP Address and Port will be determined
    by splitting the string given by the 'bind' argument.

    Otherwise if a str is passed and it does not contain the ':'
    character, a file path is assumed and a UNIXServer is created and
    bound to the file given by the 'bind' argument.
    """

    channel = 'web'

    def __init__(
        self,
        bind,
        encoding='utf-8',
        secure=False,
        certfile=None,
        channel=channel,
        display_banner=True,
        bufsize=BUFSIZE,
        **kwargs,
    ):
        """x.__init__(...) initializes x; see x.__class__.__doc__ for signature"""
        super().__init__(channel=channel)

        self._display_banner = display_banner

        SocketType = TCPServer if isinstance(bind, (int, list, tuple)) else TCPServer if ':' in bind else UNIXServer

        self.server = SocketType(
            bind,
            secure=secure,
            certfile=certfile,
            channel=channel,
            bufsize=bufsize,
            **kwargs,
        ).register(self)

        self.http = HTTP(
            self,
            encoding=encoding,
            channel=channel,
        ).register(self)

    @property
    def host(self):
        if hasattr(self, 'server'):
            return self.server.host
        return None

    @property
    def port(self):
        if hasattr(self, 'server'):
            return self.server.port
        return None

    @property
    def display_banner(self):
        return getattr(self, '_display_banner', False)

    @property
    def secure(self):
        if hasattr(self, 'server'):
            return self.server.secure
        return None

    @handler('connect')
    def _on_connect(self, *args, **kwargs):
        """Dummy Event Handler for connect"""

    @handler('closed')
    def _on_closed(self, *args, **kwargs):
        """Dummy Event Handler for closed"""

    @handler('signal')
    def _on_signal(self, *args, **kwargs):
        """Signal Event Handler"""
        self.fire(close())
        Timer(3, terminate()).register(self)

    @handler('terminate')
    def _on_terminate(self):
        raise SystemExit(0)

    @handler('ready')
    def _on_ready(self, server, bind):
        stderr.write(
            f'{self.http.version} ready! Listening on: {self.http.base}\n',
        )


class Server(BaseServer):
    """
    Create a Web Server

    Create a Web Server (HTTP) complete with the default Dispatcher to
    parse requests and posted form data dispatching to appropriate
    Controller(s).

    See: circuits.web.servers.BaseServer
    """

    def __init__(self, bind, **kwargs):
        """x.__init__(...) initializes x; see x.__class__.__doc__ for signature"""
        super().__init__(bind, **kwargs)

        Dispatcher(channel=self.channel).register(self.http)


class FakeSock:
    def getpeername(self):
        return (None, None)


class StdinServer(BaseComponent):
    channel = 'web'

    def __init__(self, encoding='utf-8', channel=channel):
        super().__init__(channel=channel)

        self.server = (io.stdin + io.stdout).register(self)
        self.http = HTTP(
            self,
            encoding=encoding,
            channel=channel,
        ).register(self)

        Dispatcher(channel=self.channel).register(self)

    @property
    def host(self):
        return io.stdin.filename

    @property
    def port(self):
        return 0

    @property
    def secure(self):
        return False

    @handler('read', channel='stdin')
    def read(self, data):
        self.fire(read(FakeSock(), data))

    @handler('write')
    def write(self, sock, data):
        self.fire(write(data))



"""
VirtualHost

This module implements a virtual host dispatcher that sends requests
for configured virtual hosts to different dispatchers.
"""

from urllib.parse import urljoin

from circuits import BaseComponent, handler


class VirtualHosts(BaseComponent):
    """
    Forward to anotehr Dispatcher based on the Host header.

    This can be useful when running multiple sites within one server.
    It allows several domains to point to different parts of a single
    website structure. For example:
    - http://www.domain.example      -> /
    - http://www.domain2.example     -> /domain2
    - http://www.domain2.example:443 -> /secure

    :param domains: a dict of {host header value: virtual prefix} pairs.
    :type  domains: dict

    The incoming "Host" request header is looked up in this dict,
    and, if a match is found, the corresponding "virtual prefix"
    value will be prepended to the URL path before passing the
    request onto the next dispatcher.

    Note that you often need separate entries for "example.com"
    and "www.example.com". In addition, "Host" headers may contain
    the port number.
    """

    channel = 'web'

    def __init__(self, domains):
        super().__init__()

        self.domains = domains

    @handler('request', priority=1.0)
    def _on_request(self, event, request, response):
        path = request.path.strip('/')

        header = request.headers.get
        domain = header('X-Forwarded-Host', header('Host', ''))
        prefix = self.domains.get(domain, '')

        if prefix:
            path = urljoin('/%s/' % prefix, path)
            request.path = path



"""
Request/Response Wrappers

This module implements the Request and Response objects.
"""

from email.utils import formatdate
from functools import partial
from http.cookies import SimpleCookie
from io import BytesIO
from time import time

from circuits.net.sockets import BUFSIZE

from .constants import HTTP_STATUS_CODES, SERVER_VERSION
from .errors import httperror
from .headers import Headers
from .url import parse_url


formatdate = partial(formatdate, usegmt=True)


def file_generator(input, chunkSize=BUFSIZE):
    chunk = input.read(chunkSize)
    while chunk:
        yield chunk
        chunk = input.read(chunkSize)
    input.close()


class Host:
    """
    An internet address.

    name should be the client's host name. If not available (because no DNS
    lookup is performed), the IP address should be used instead.
    """

    ip = '0.0.0.0'
    port = 80
    name = 'unknown.tld'

    def __init__(self, ip, port, name=None):
        self.ip = ip
        self.port = port
        if name is None:
            name = ip
        self.name = name

    def __repr__(self):
        return f'Host({self.ip!r}, {self.port!r}, {self.name!r})'


class HTTPStatus:
    __slots__ = ('_reason', '_status')

    def __init__(self, status=200, reason=None):
        self._status = status
        self._reason = reason or HTTP_STATUS_CODES.get(status, '')

    def __int__(self):
        return self._status

    def __lt__(self, other):
        if isinstance(other, int):
            return self._status < other
        return super().__lt__(other)

    def __gt__(self, other):
        if isinstance(other, int):
            return self._status > other
        return super().__gt__(other)

    def __le__(self, other):
        if isinstance(other, int):
            return self._status <= other
        return super().__le__(other)

    def __ge__(self, other):
        if isinstance(other, int):
            return self._status >= other
        return super().__ge__(other)

    def __eq__(self, other):
        if isinstance(other, int):
            return self._status == other
        return super().__eq__(other)

    def __str__(self):
        return f'{self._status:d} {self._reason}'

    def __repr__(self):
        return f'<Status (status={self._status:d} reason={self._reason}>'

    def __format__(self, format_spec):
        return format(str(self), format_spec)

    @property
    def status(self):
        return self._status

    @property
    def reason(self):
        return self._reason


class Request:
    """
    Creates a new Request object to hold information about a request.

    :param sock: The socket object of the request.
    :type  sock: socket.socket

    :param method: The requested method.
    :type  method: str

    :param scheme: The requested scheme.
    :type  scheme: str

    :param path: The requested path.
    :type  path: str

    :param protocol: The requested protocol.
    :type  protocol: str

    :param qs: The query string of the request.
    :type  qs: str
    """

    server = None
    """:cvar: A reference to the underlying server"""

    scheme = 'http'
    protocol = (1, 1)
    host = ''
    local = Host('127.0.0.1', 80)
    remote = Host('', 0)

    index = None
    script_name = ''

    login = None
    handled = False

    def __init__(self, sock, method='GET', scheme='http', path='/', protocol=(1, 1), qs='', headers=None, server=None):
        """Initializes x; see x.__class__.__doc__ for signature"""
        self.sock = sock
        self.method = method
        self.scheme = scheme or Request.scheme
        self.path = path
        self.protocol = protocol
        self.qs = qs
        self.print_debug = getattr(server, 'display_banner', False)

        self.headers = headers or Headers()
        self.server = server

        self.cookie = SimpleCookie()

        if sock is not None:
            name = sock.getpeername()
            try:
                ip, port = name
                name = None
            except ValueError:  # AF_UNIX
                ip, port = None, None
            self.remote = Host(ip, port, name)

        cookie = self.headers.get('Cookie')
        if cookie is not None:
            self.cookie.load(cookie)

        self.body = BytesIO()

        if self.server is not None:
            self.local = Host(self.server.host, self.server.port)

        try:
            host = self.headers['Host']
            if ':' in host:
                parts = host.split(':', 1)
                host = parts[0]
                port = int(parts[1])
            else:
                port = 443 if self.scheme == 'https' else 80
        except KeyError:
            host = self.local.name or self.local.ip
            port = self.server.port

        self.host = host
        self.port = port

        base = '{}://{}{}/'.format(
            self.scheme,
            self.host,
            f':{self.port:d}' if self.port not in (80, 443) else '',
        )

        self.base = parse_url(base)

        url = '{}{}{}'.format(
            base,
            self.path,
            f'?{self.qs}' if self.qs else '',
        )
        self.uri = parse_url(url)
        self.uri.sanitize()

    def __repr__(self):
        protocol = 'HTTP/%d.%d' % self.protocol
        return f'<Request {self.method} {self.path} {protocol}>'


class Body:
    """Response Body"""

    encode_errors = 'strict'

    def __get__(self, response, cls=None):
        if response is None:
            return self
        return response._body

    def __set__(self, response, value):
        if response == value:
            return

        if isinstance(value, bytes):
            value = [value] if value else []
        elif isinstance(value, str):
            value = [value.encode(response.encoding, self.encode_errors)] if value else []
        elif hasattr(value, 'read'):
            response.stream = True
            value = file_generator(value)
        elif isinstance(value, httperror):
            value = [str(value)]
        elif value is None:
            value = []

        response._body = value


class Status:
    """Response Status"""

    def __get__(self, response, cls=None):
        if response is None:
            return self
        return response._status

    def __set__(self, response, value):
        value = HTTPStatus(value) if isinstance(value, int) else value

        response._status = value


class Response:
    """
    Response(sock, request) -> new Response object

    A Response object that holds the response to
    send back to the client. This ensure that the correct data
    is sent in the correct order.
    """

    body = Body()
    status = Status()

    done = False
    close = False
    stream = False
    chunked = False

    def __init__(self, request, encoding='utf-8', status=None):
        """Initializes x; see x.__class__.__doc__ for signature"""
        self.request = request
        self.encoding = encoding

        self._body = []
        self._status = HTTPStatus(status if status is not None else 200)

        self.time = time()

        self.headers = Headers()
        self.headers['Date'] = formatdate()

        if getattr(self.request.server, 'display_banner', False):
            if self.request.server is not None:
                self.headers.add_header('Server', request.server.http.version)
            else:
                self.headers.add_header('X-Powered-By', SERVER_VERSION)

        self.cookie = self.request.cookie

        self.protocol = 'HTTP/%d.%d' % self.request.protocol

    def __repr__(self):
        return '<Response %s %s (%d)>' % (
            self.status,
            self.headers.get('Content-Type'),
            (len(self.body) if isinstance(self.body, str) else 0),
        )

    def __str__(self):
        return f'{self.protocol} {self.status}\r\n'

    def __bytes__(self):
        return str(self).encode('ISO8859-1')

    def prepare(self):
        # Set a default content-Type if we don't have one.
        self.headers.setdefault(
            'Content-Type',
            f'text/html; charset={self.encoding}',
        )

        cLength = None
        if self.body is not None:
            if isinstance(self.body, bytes):
                cLength = len(self.body)
            elif isinstance(self.body, str):
                cLength = len(self.body.encode(self.encoding))
            elif isinstance(self.body, list):
                cLength = sum(
                    len(s.encode(self.encoding)) if not isinstance(s, bytes) else len(s) for s in self.body if s is not None
                )

        if cLength is not None:
            self.headers['Content-Length'] = str(cLength)

        for v in self.cookie.values():
            self.headers.add_header('Set-Cookie', v.OutputString())

        status = self.status

        if status == 413:
            self.close = True
        elif 'Content-Length' not in self.headers:
            if status < 200 or status in (204, 205, 304):
                pass
            else:
                if (
                    self.protocol == 'HTTP/1.1'
                    and self.request.method != 'HEAD'
                    and self.request.server is not None
                    and cLength != 0
                ):
                    self.chunked = True
                    self.headers.add_header('Transfer-Encoding', 'chunked')
                else:
                    self.close = True

        if self.request.server is not None and 'Connection' not in self.headers:
            if self.protocol == 'HTTP/1.1':
                if self.close:
                    self.headers.add_header('Connection', 'close')
            else:
                if not self.close:
                    self.headers.add_header('Connection', 'Keep-Alive')

        if self.headers.get('Transfer-Encoding', '') == 'chunked':
            self.chunked = True



#!/usr/bin/env python
from io import BytesIO

import pytest

from circuits import Component
from circuits.io import File
from circuits.io.events import close, write


pytestmark = pytest.mark.skipif(pytest.PLATFORM == 'win32', reason='Unsupported Platform')


class FileApp(Component):
    def init(self, *args, **kwargs):
        self.file = File(*args, **kwargs).register(self)

        self.eof = False
        self.closed = False
        self.buffer = BytesIO()

    def read(self, data):
        self.buffer.write(data)

    def eof(self):
        self.eof = True

    def closed(self):
        self.closed = True


def test_open_fileobj(manager, watcher, tmpdir):
    filename = str(tmpdir.ensure('helloworld.txt'))
    with open(filename, 'w') as f:
        f.write('Hello World!')

    fileobj = open(filename)

    app = FileApp(fileobj).register(manager)
    assert watcher.wait('opened', app.file.channel)

    assert watcher.wait('eof', app.file.channel)
    assert app.eof

    app.fire(close(), app.file.channel)
    assert watcher.wait('closed', app.file.channel)
    assert app.closed

    app.unregister()
    assert watcher.wait('unregistered')

    s = app.buffer.getvalue()
    assert s == b'Hello World!'


def test_read_write(manager, watcher, tmpdir):
    filename = str(tmpdir.ensure('helloworld.txt'))

    app = FileApp(filename, 'w').register(manager)
    assert watcher.wait('opened', app.file.channel)

    app.fire(write(b'Hello World!'), app.file.channel)
    assert watcher.wait('write', app.file.channel)

    app.fire(close(), app.file.channel)
    assert watcher.wait('closed', app.file.channel)
    assert app.closed

    app.unregister()
    assert watcher.wait('unregistered')

    app = FileApp(filename, 'r').register(manager)
    assert watcher.wait('opened', app.file.channel)

    assert watcher.wait('eof', app.file.channel)
    assert app.eof

    app.fire(close(), app.file.channel)
    assert watcher.wait('closed', app.file.channel)
    assert app.closed

    app.unregister()
    assert watcher.wait('unregistered')

    s = app.buffer.getvalue()
    assert s == b'Hello World!'



"""
Macro

Macro support and dispatcher
"""

import os
from inspect import getmembers, getmodule, isfunction

from creoleparser import parse_args


class Macro:
    def __init__(self, name, arg_string, body, isblock):
        super().__init__()

        self.name = name
        self.arg_string = arg_string
        self.body = body
        self.isblock = isblock


def dispatcher(name, arg_string, body, isblock, environ):
    if name in environ['macros']:
        macro = Macro(name, arg_string, body, isblock)
        args, kwargs = parse_args(arg_string)
        try:
            return environ['macros'][name](macro, environ, *args, **kwargs)
        except Exception as e:
            return f'ERROR: Error while executing macro {name!r} ({e})'
    else:
        return 'Macro not found!'


def loadMacros():
    path = os.path.abspath(os.path.dirname(__file__))

    def p(x):
        return os.path.splitext(x)[1] == '.py'

    modules = [x for x in os.listdir(path) if p(x) and x != '__init__.py']

    macros = {}

    for module in modules:
        name, _ = os.path.splitext(module)

        moduleName = f'{__package__}.{name}'
        m = __import__(moduleName, globals(), locals(), __package__)

        def p(x):
            return isfunction(x) and getmodule(x) is m

        for name, function in getmembers(m, p):
            name = name.replace('_', '-')
            try:
                macros[name] = function
            except Exception:
                continue

    return macros



"""
Circuits Tools

circuits.tools contains a standard set of tools for circuits. These
tools are installed as executables with a prefix of "circuits."
"""

import inspect as _inspect
from functools import wraps
from warnings import warn, warn_explicit


def tryimport(modules, obj=None, message=None):
    modules = (modules,) if isinstance(modules, str) else modules

    for module in modules:
        try:
            m = __import__(module, globals(), locals())
            return getattr(m, obj) if obj is not None else m
        except Exception:
            pass

    if message is not None:
        warn(message)
        return None
    return None


def getargspec(func):
    getargs = _inspect.getfullargspec if hasattr(_inspect, 'getfullargspec') else _inspect.getargspec
    return getargs(func)[:4]


def walk(x, f, d=0, v=None):
    if not v:
        v = set()
    yield f(d, x)
    for c in x.components.copy():
        if c not in v:
            v.add(c)
            yield from walk(c, f, d + 1, v)


def edges(x, e=None, v=None, d=0):
    if not e:
        e = set()
    if not v:
        v = []
    d += 1
    for c in x.components.copy():
        e.add((x, c, d))
        edges(c, e, v, d)
    return e


def findroot(x):
    if x.parent == x:
        return x
    return findroot(x.parent)


def kill(x):
    for c in x.components.copy():
        kill(c)
    if x.parent is not x:
        x.unregister()


def _graph(x):
    """
    Create a directed graph of the Component structure of x

    :param x: A Component or Manager to graph
    :type  x: Component or Manager
    """
    networkx = tryimport('networkx')
    pygraphviz = tryimport('pygraphviz')
    plt = tryimport('matplotlib.pyplot', 'pyplot')

    if not all([networkx, pygraphviz, plt]):
        return None, None

    graph_edges = []
    for u, v, d in edges(x):
        graph_edges.append((u.name, v.name, float(d)))

    g = networkx.DiGraph()
    g.add_weighted_edges_from(graph_edges)

    elarge = [(u, v) for (u, v, d) in g.edges(data=True) if d['weight'] > 3.0]
    esmall = [(u, v) for (u, v, d) in g.edges(data=True) if d['weight'] <= 3.0]

    pos = networkx.spring_layout(g)  # positions for all nodes

    # nodes
    networkx.draw_networkx_nodes(g, pos, node_size=700)

    # edges
    networkx.draw_networkx_edges(g, pos, edgelist=elarge, width=1)
    networkx.draw_networkx_edges(
        g,
        pos,
        edgelist=esmall,
        width=1,
        alpha=0.5,
        edge_color='b',
        style='dashed',
    )

    # labels
    networkx.draw_networkx_labels(
        g,
        pos,
        font_size=10,
        font_family='sans-serif',
    )

    plt.axis('off')

    plt.clf()
    return plt, g


def graph_ascii(x):
    """
    Display a directed graph of the Component structure of x

    :param x: A Component or Manager to graph
    :type  x: Component or Manager

    @return: A directed graph representing x's Component structure.
    @rtype:  str
    """

    def printer(d, x):
        return '%s* %s' % (' ' * d, x)

    return '\n'.join(walk(x, printer))


def graph_dot(x, name=None):
    """
    :param x: A Component or Manager to graph
    :type  x: Component or Manager

    :param name: A name for the graph (defaults to x's name)
    :type  name: str
    """
    networkx = tryimport('networkx')
    _plt, g = _graph(x)
    if g is not None:
        networkx.drawing.nx_agraph.write_dot(g, f'{name or x.name}.dot')


def graph_png(x, name=None):
    """
    :param x: A Component or Manager to graph
    :type  x: Component or Manager

    :param name: A name for the graph (defaults to x's name)
    :type  name: str
    """
    plt, _g = _graph(x)
    if plt is not None:
        plt.savefig(f'{name or x.name}.png')


def graph(x, name=None):
    """
    :param x: A Component or Manager to graph
    :type  x: Component or Manager

    :param name: A name for the graph (defaults to x's name)
    :type  name: str
    """
    graph_dot(x, name)
    graph_png(x, name)
    return graph_ascii(x)


def inspect(x):
    """
    Display an inspection report of the Component or Manager x

    :param x: A Component or Manager to graph
    :type  x: Component or Manager

    @return: A detailed inspection report of x
    @rtype:  str
    """
    s = []
    write = s.append

    write(' Components: %d\n' % len(x.components))
    for component in x.components:
        write('  %s\n' % component)
    write('\n')

    from circuits import reprhandler

    write(' Event Handlers: %d\n' % len(x._handlers.values()))
    for event in x._handlers:
        write('  %s; %d\n' % (event, len(x._handlers[event])))
        for handler in x._handlers[event]:
            write('   %s\n' % reprhandler(handler))

    return ''.join(s)


def deprecated(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        warn_explicit(
            f'Call to deprecated function {f.__name__}',
            category=DeprecationWarning,
            filename=f.__code__.co_filename,
            lineno=f.__code__.co_firstlineno + 1,
        )
        return f(*args, **kwargs)

    return wrapper



from circuits.web import Controller

from .helpers import HTTPError, urlopen


class Root(Controller):
    def index(self):
        pass


def test(webapp):
    try:
        urlopen(webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == 'Not Found'
    else:
        assert False



#!/usr/bin/env python
from http.client import HTTPConnection

import pytest

from circuits.web import Controller


class Root(Controller):
    def GET(self):
        return 'GET'

    def PUT(self):
        return 'PUT'

    def POST(self):
        return 'POST'

    def DELETE(self):
        return 'DELETE'


@pytest.mark.parametrize(('method'), ['GET', 'PUT', 'POST', 'DELETE'])
def test(webapp, method):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    connection.request(method, '/')
    response = connection.getresponse()
    assert response.status == 200
    assert response.reason == 'OK'

    s = response.read()
    assert s == f'{method:s}'.encode()

    connection.close()



"""Internet Relay Chat Protocol replies"""

from .message import Message


def _M(*args, **kwargs):
    kwargs['add_nick'] = True
    return Message(*args, **kwargs)


def ERROR(host, reason=None):
    return Message('ERROR', ':Closing link: {} ({})'.format(host, reason or ''))


def JOIN(name, prefix=None):
    return Message('JOIN', name, prefix=prefix)


def KICK(channel, nick, reason=None, prefix=None):
    return Message('KICK', channel, nick, reason, prefix=prefix)


def MODE(target, modes, params=None, prefix=None):
    if params is None:
        return Message('MODE', target, modes, prefix=prefix)
    return Message('MODE', target, modes, ' '.join(params), prefix=prefix)


def PART(channel, nick, reason=None, prefix=None):
    return Message('PART', channel, nick, reason, prefix=prefix)


def PING(server):
    return Message('PING', f':{server}')


def PONG(server, text):
    return Message('PONG', server, f':{text}')


def TOPIC(channel, topic, prefix=None):
    return Message('TOPIC', channel, topic, prefix=prefix)


def RPL_WELCOME(network):
    return _M('001', f'Welcome to the {network} IRC Network')


def RPL_YOURHOST(host, version):
    return _M('002', f'Your host is {host} running {version}')


def RPL_CREATED(date):
    return _M('003', f'This server was created {date}')


def RPL_MYINFO(server, version, umodes, chmodes):
    return _M('004', server, version, umodes, chmodes)


def RPL_ISUPPORT(features):
    return _M('005', *(features + ('are supported by this server',)))


def RPL_UMODEIS(modes):
    return _M('221', modes)


def RPL_LUSERCLIENT(nusers, nservices, nservers):
    return _M(
        '251',
        f'There are {nusers} users and {nservices} services on {nservers} servers',
    )


def RPL_LUSEROP(noperators):
    return _M('252', f'{noperators}', 'operator(s) online')


def RPL_LUSERUNKOWN(nunknown):
    return _M('253', f'{nunknown}', 'unknown connection(s)')


def RPL_LUSERCHANNELS(nchannels):
    return _M('254', f'{nchannels}', 'channels formed')


def RPL_LUSERME(nclients, nservers):
    return _M('255', f'I have {nclients} clients and {nservers} servers')


def RPL_AWAY(nick, message):
    return _M('301', nick, f':{message}')


def RPL_UNAWAY():
    return _M('305', 'You are no longer marked as being away')


def RPL_NOWAWAY():
    return _M('306', 'You have been marked as being away')


def RPL_WHOISUSER(nick, user, host, realname):
    return _M('311', nick, user, host, '*', f':{realname}')


def RPL_WHOISSERVER(nick, server, server_info):
    return _M('312', nick, server, server_info)


def RPL_WHOISOPERATOR(nick):
    return _M('313', nick, 'is an IRC operator')


def RPL_ENDOFWHO(mask):
    return _M('315', mask, 'End of WHO list')


def RPL_WHOISIDLE(nick, idle, signon):
    return _M(
        '317',
        nick,
        f'{idle}',
        f'{signon}',
        'seconds idle, signon time',
    )


def RPL_ENDOFWHOIS(nick):
    return _M('318', nick, 'End of WHOIS list')


def RPL_WHOISCHANNELS(nick, channels):
    return _M('319', nick, ':{}'.format(' '.join(channels)))


def RPL_LISTSTART(header=None):
    return _M('321', header or 'Channels :Users Name')


def RPL_LIST(channel, nvisible, topic):
    return _M('322', channel, f'{nvisible}', topic)


def RPL_LISTEND():
    return _M('323', 'End of LIST')


def RPL_CHANNELMODEIS(channel, mode, params=None):
    if params is None:
        return _M('324', channel, mode)
    return _M('324', channel, mode, params)


def RPL_NOTOPIC(channel):
    return _M('331', channel, 'No topic is set')


def RPL_TOPIC(channel, topic):
    return _M('332', channel, topic)


def RPL_TOPICWHO(channel, setter, timestamp):
    return _M('333', channel, setter, f'{timestamp}')


def RPL_INVITING(channel, nick):
    return _M('341', f'{channel} {nick}')


def RPL_SUMMONING(user):
    return _M('342', f'{user} :Summoning user to IRC')


def RPL_INVITELIST(channel, invitemask):
    return _M('346', f'{channel} {invitemask}')


def RPL_ENDOFINVITELIST(channel):
    return _M('347', f'{channel} :End of channel invite list')


def RPL_VERSION(name, version, hostname, url):
    return _M('351', name, version, hostname, url)


def RPL_WHOREPLY(channel, user, host, server, nick, status, hops, name):
    return _M(
        '352',
        channel,
        user,
        host,
        server,
        nick,
        status,
        f':{hops} {name}',
    )


def RPL_NAMEREPLY(channel, names):
    return _M('353', '=', channel, ' '.join(names))


def RPL_ENDOFNAMES(channel):
    return _M('366', channel, 'End of NAMES list')


def RPL_MOTD(text):
    return _M('372', f'- {text}')


def RPL_MOTDSTART(server):
    return _M('375', f'- {server} Message of the day -')


def RPL_ENDOFMOTD():
    return _M('376', 'End of MOTD command')


def RPL_YOUREOPER():
    return _M('381', 'You are now an IRC operator')


def ERR_NOSUCHNICK(nick):
    return _M('401', nick, 'No such nick')


def ERR_NOSUCHCHANNEL(channel):
    return _M('403', channel, 'No such channel')


def ERR_CANNOTSENDTOCHAN(channel):
    return _M('404', channel, 'Cannot send to channel')


def ERR_TOOMANYCHANNELS(channel):
    return _M('405', channel, 'You have joined too many channels')


def ERR_UNKNOWNCOMMAND(command):
    return _M('421', command, 'Unknown command')


def ERR_NOMOTD():
    return _M('422', 'MOTD file is missing')


def ERR_NONICKNAMEGIVEN():
    return _M('431', 'No nickname given')


def ERR_ERRONEUSNICKNAME(nick):
    return _M('432', nick, 'Erroneous nickname')


def ERR_NICKNAMEINUSE(nick):
    return _M('433', nick, 'Nickname is already in use')


def ERR_USERNOTINCHANNEL(nick, channel):
    return _M('441', nick, channel, "They aren't on that channel")


def ERR_NOTREGISTERED():
    return _M('451', 'You have not registered')


def ERR_NEEDMOREPARAMS(command):
    return _M('461', command, 'Need more parameters')


def ERR_PASSWDMISMATCH():
    return _M('464', 'Password incorrect')


def ERR_UNKNOWNMODE(mode, channel=None):
    if channel is None:
        return _M('472', mode, 'is unknown mode char to me')
    return _M('472', mode, f'is unknown mode char to me for channel {channel}')


def ERR_CHANOPRIVSNEEDED(channel):
    return _M('482', channel, "You're not channel operator")


def ERR_NOPRIVILEGES():
    return _M('481', "Permission Denied- You're not an IRC operator")


def ERR_NOOPERHOST():
    return _M('491', 'No O-lines for your host')


def ERR_USERSDONTMATCH():
    return _M('502', 'Cannot change mode for other users')



#!/usr/bin/env python
from circuits.web import Controller, Sessions

from .helpers import CookieJar, HTTPCookieProcessor, build_opener


class Root(Controller):
    def index(self, vpath=None):
        if vpath:
            name = vpath
            with self.session as data:
                data['name'] = name
        else:
            name = self.session.get('name', 'World!')

        return 'Hello %s' % name

    def logout(self):
        self.session.expire()
        return 'OK'


def test(webapp):
    Sessions().register(webapp)

    cj = CookieJar()
    opener = build_opener(HTTPCookieProcessor(cj))

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    f = opener.open(webapp.server.http.base + '/test')
    s = f.read()
    assert s == b'Hello test'

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello test'


def test_expire(webapp):
    Sessions().register(webapp)

    cj = CookieJar()
    opener = build_opener(HTTPCookieProcessor(cj))

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    f = opener.open(webapp.server.http.base + '/test')
    s = f.read()
    assert s == b'Hello test'

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello test'

    f = opener.open(webapp.server.http.base + '/logout')
    s = f.read()
    assert s == b'OK'

    f = opener.open(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'



#!/usr/bin/env python
import json

import pytest

from circuits.web import Controller
from circuits.web.wsgi import Application

from .helpers import HTTPError, urlencode, urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'

    def test_args(self, *args, **kwargs):
        self.response.headers['Content-Type'] = 'application/json'
        return json.dumps(
            {
                'args': args,
                'kwargs': kwargs,
                'path': self.request.path,
                'uri_path': self.request.uri._path.decode(),
                'base_path': self.request.base._path.decode(),
                'method': self.request.method,
                'scheme': self.request.scheme,
                'protocol': self.request.protocol,
                'qs': self.request.qs,
                'script_name': self.request.script_name,
                'content_type': self.request.headers['Content-Type'],
            }
        )

    def test_redirect(self):
        return self.redirect('/')

    def test_forbidden(self):
        return self.forbidden()

    def test_notfound(self):
        return self.notfound()


application = Application() + Root()


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_404(webapp):
    with pytest.raises(HTTPError) as exc:
        urlopen('%s/foo' % webapp.server.http.base)
    assert exc.value.code == 404
    assert exc.value.msg == 'Not Found'


def test_args(webapp):
    args = ['1', '2', '3']
    kwargs = {'1': 'one', '2': 'two', '3': 'three'}
    url = '%s/test_args/%s' % (webapp.server.http.base, '/'.join(args))
    data = urlencode(kwargs).encode()

    f = urlopen(url, data)
    data = json.load(f)
    assert data['args'] == args
    assert data['kwargs'] == kwargs
    assert data['path'] == 'test_args/1/2/3'
    assert data['uri_path'] == '/test_args/1/2/3'
    assert data['base_path'] == '/'
    assert data['method'] == 'POST'
    assert data['scheme'] == 'http'
    assert data['protocol'] == [1, 1]
    assert data['qs'] == ''
    assert data['script_name'] == '/'
    assert data['content_type'] == 'application/x-www-form-urlencoded'


def test_redirect(webapp):
    f = urlopen('%s/test_redirect' % webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_forbidden(webapp):
    with pytest.raises(HTTPError) as exc:
        urlopen('%s/test_forbidden' % webapp.server.http.base)
    assert exc.value.code == 403
    assert exc.value.msg == 'Forbidden'


def test_notfound(webapp):
    with pytest.raises(HTTPError) as exc:
        urlopen('%s/test_notfound' % webapp.server.http.base)
    assert exc.value.code == 404
    assert exc.value.msg == 'Not Found'



"""
This module implements a generic Loader suitable for dynamically loading
components from other modules. This supports loading from local paths,
eggs and zip archives. Both setuptools and distribute are fully supported.
"""

import sys
from inspect import getmembers, getmodule, isclass

from .components import BaseComponent
from .handlers import handler
from .utils import safeimport


class Loader(BaseComponent):
    """
    Create a new Loader Component

    Creates a new Loader Component that enables dynamic loading of
    components from modules either in local paths, eggs or zip archives.
    """

    channel = 'loader'

    def __init__(self, auto_register=True, init_args=None, init_kwargs=None, paths=None, channel=channel):
        """Initializes x; see x.__class__.__doc__ for signature"""
        super().__init__(channel=channel)

        self._auto_register = auto_register
        self._init_args = init_args or ()
        self._init_kwargs = init_kwargs or {}

        if paths:
            for path in paths:
                sys.path.insert(0, path)

    @handler('load')
    def load(self, name):
        module = safeimport(name)
        if module is not None:

            def test(x):
                return isclass(x) and issubclass(x, BaseComponent) and getmodule(x) is module

            components = [x[1] for x in getmembers(module, test)]

            if components:
                TheComponent = components[0]

                component = TheComponent(*self._init_args, **self._init_kwargs)

                if self._auto_register:
                    component.register(self)

                return component
            return None
        return None



"""
Bridge

The Bridge Component is used for inter-process communications between
processes. Bridge is used internally when a Component is started in
"process mode" via :meth:`circuits.core.manager.start`. Typically a
Pipe is used as the socket transport between two sides of a Bridge
(*there must be a :class:`~Bridge` instance on both sides*).
"""

import traceback
from pickle import dumps, loads

from .components import BaseComponent
from .events import Event, exception
from .handlers import handler
from .values import Value


_sentinel = b'~~~'


class ipc(Event):
    """
    ipc Event

    Send an event to a child/parent process
    """

    def __init__(self, event, channel=None):
        """
        :param event:   Event to execute remotely.
        :type event:    :class:`circuits.core.events.Event`

        :param channel: IPC Channel (channel to use on child/parent).
        :type channel:  str
        """
        super().__init__(event, channel=channel)


class Bridge(BaseComponent):
    channel = 'bridge'

    def init(self, socket, channel=channel):
        self._buffer = b''
        self._socket = socket
        self._values = {}

        if self._socket is not None:
            self._socket.register(self)

    def _process_packet(self, eid, obj):
        if isinstance(obj, Event):
            obj.remote = True
            obj.notify = 'value_changed'
            obj.waitingHandlers = 0
            value = self.fire(obj)
            self._values[value] = eid
        elif isinstance(obj, Value):
            if obj.result:
                if isinstance(obj.value, list):
                    for item in obj.value:
                        self._values[eid].value = item
                else:
                    self._values[eid].value = obj.value
            event = Event.create(Bridge.__waiting_event(eid))
            event.remote = True
            self.fire(event, self.channel)

    @handler('value_changed', channel='*')
    def _on_value_changed(self, value):
        try:
            eid = self._values[value]
            if value.errors:
                Bridge.__adapt_error_value(value)
            self.__write(eid, value)
        except Exception:
            pass

    @handler('read')
    def _on_read(self, data):
        self._buffer += data
        items = self._buffer.split(_sentinel)

        if items[-1] != '':
            self._buffer = items.pop()

        for item in filter(None, items):
            self._process_packet(*loads(item))

    def __send(self, eid, event):
        try:
            if isinstance(event, exception):
                Bridge.__adapt_exception(event)
            self._values[eid] = event.value
            self.__write(eid, event)
        except Exception:
            pass

    def __write(self, eid, data):
        self._socket.write(dumps((eid, data)) + _sentinel)

    @handler('ipc')
    def _on_ipc(self, event, ipc_event, channel=None):
        """
        Send event to a child/parentprocess

        Event handler to run an event on a child/parent process
        (the event definition is :class:`circuits.core.bridge.ipc`)

        :param event:    The event triggered (by the handler)
        :type event:     :class:`circuits.node.events.remote`

        :param ipc_event:    Event to execute in child/parent process.
        :type ipc_event:     :class:`circuits.core.events.Event`

        :param channel:    Remote channel (channel to use on peer).
        :type channel:     str

        :return: The result of remote event
        :rtype: generator

        :Example:
        ``# hello is your event to execute in the child process
        result = yield self.fire(ipc(hello()))
        print(result.value)``
        """
        ipc_event.channels = (channel,) if channel is not None else event.channels
        event.value.value = ipc_event.value = Value(ipc_event, self)

        eid = hash(ipc_event)
        self.__send(eid, ipc_event)
        yield self.wait(Bridge.__waiting_event(eid))

    @staticmethod
    def __waiting_event(eid):
        return '%s_done' % eid

    @staticmethod
    def __adapt_exception(ex):
        fevent_value = ex.kwargs['fevent'].value
        Bridge.__adapt_error_value(fevent_value)

    @staticmethod
    def __adapt_error_value(value):
        if not isinstance(value[2], list):
            value._value = (value[0], value[1], traceback.extract_tb(value[2]))



#!/usr/bin/env python

from time import sleep

from circuits import Component, Debugger, Event, Timer, Worker, task


def factorial(n):
    x = 1
    for i in range(1, (n + 1)):
        x = x * (i + 1)
        sleep(1)  # deliberate!
    return x


class App(Component):
    def init(self, *args, **kwargs):
        Worker(process=True).register(self)

    def foo(self):
        print('Foo!')

    def started(self, component):
        Timer(1, Event.create('foo'), persist=True).register(self)
        x = yield self.call(task(factorial, 10))
        print(f'{x.value:d}')
        self.stop()


app = App()
Debugger().register(app)
app.run()



#
# This file is part of http-parser released under the MIT license.
# See the NOTICE for more information.
#
# This module is liberally borrowed (with modifications) from:
# https://raw.githubusercontent.com/benoitc/http-parser/master/http_parser/pyparser.py
import contextlib
import re
import zlib
from sys import maxsize
from urllib.parse import urlsplit

from circuits.web.headers import Headers


METHOD_RE = re.compile('^[A-Z0-9$-_.]{1,20}$')
VERSION_RE = re.compile(r'^HTTP/(\d+).(\d+)$')
STATUS_RE = re.compile(r'^(\d{3})(?:\s+([\s\w]*))$')
HEADER_RE = re.compile('[\\x00-\\x1F\\x7F()<>@,;:/\\[\\]={} \\t\\\\"]')

# errors
BAD_FIRST_LINE = 0
INVALID_HEADER = 1
INVALID_CHUNK = 2


class InvalidRequestLine(Exception):
    """error raised when first line is invalid"""


class InvalidHeader(Exception):
    """error raised on invalid header"""


class InvalidChunkSize(Exception):
    """error raised when we parse an invalid chunk size"""


class HttpParser:
    def __init__(self, kind=2, decompress=False):
        self.kind = kind
        self.decompress = decompress

        # errors vars
        self.errno = None
        self.errstr = ''

        # protected variables
        self._buf = []
        self._version = None
        self._method = None
        self._status_code = None
        self._status = None
        self._reason = None
        self._url = None
        self._path = None
        self._query_string = None
        self._headers = Headers([])
        self._environ = {}
        self._chunked = False
        self._body = []
        self._trailers = None
        self._partial_body = False
        self._clen = None
        self._clen_rest = None

        # private events
        self.__on_firstline = False
        self.__on_headers_complete = False
        self.__on_message_begin = False
        self.__on_message_complete = False

        self.__decompress_obj = None
        self.__decompress_first_try = True

    def get_version(self):
        return self._version

    def get_method(self):
        return self._method

    def get_status_code(self):
        return self._status_code

    def get_url(self):
        return self._url

    def get_scheme(self):
        return self._scheme

    def get_path(self):
        return self._path

    def get_query_string(self):
        return self._query_string

    def get_headers(self):
        return self._headers

    def recv_body(self):
        """Return last chunk of the parsed body"""
        body = b''.join(self._body)
        self._body = []
        self._partial_body = False
        return body

    def recv_body_into(self, barray):
        """
        Receive the last chunk of the parsed body and store the data
        in a buffer rather than creating a new string.
        """
        length = len(barray)
        body = b''.join(self._body)
        m = min(len(body), length)
        data, rest = body[:m], body[m:]
        barray[0:m] = data
        if not rest:
            self._body = []
            self._partial_body = False
        else:
            self._body = [rest]
        return m

    def is_upgrade(self):
        """
        Do we get upgrade header in the request. Useful for
        websockets
        """
        hconn = self._headers.get('connection', '').lower()
        hconn_parts = [x.strip() for x in hconn.split(',')]
        return 'upgrade' in hconn_parts

    def is_headers_complete(self):
        """Return True if all headers have been parsed."""
        return self.__on_headers_complete

    def is_partial_body(self):
        """Return True if a chunk of body have been parsed"""
        return self._partial_body

    def is_message_begin(self):
        """Return True if the parsing start"""
        return self.__on_message_begin

    def is_message_complete(self):
        """Return True if the parsing is done (we get EOF)"""
        return self.__on_message_complete

    def is_chunked(self):
        """Return True if Transfer-Encoding header value is chunked"""
        return self._chunked

    def should_keep_alive(self):
        """Return True if the connection should be kept alive"""
        hconn = self._headers.get('connection', '').lower()
        if hconn == 'close':
            return False
        if hconn == 'keep-alive':
            return True
        return self._version == (1, 1)

    def execute(self, data, length):
        # end of body can be passed manually by putting a length of 0

        if length == 0:
            self.__on_message_complete = True
            return length

        # start to parse
        nb_parsed = 0
        while True:
            if not self.__on_firstline:
                idx = data.find(b'\r\n')
                if idx < 0:
                    self._buf.append(data)
                    return len(data)
                self.__on_firstline = True
                self._buf.append(data[:idx])
                first_line = b''.join(self._buf)
                first_line = str(first_line, 'unicode_escape')
                nb_parsed = nb_parsed + idx + 2

                rest = data[idx + 2 :]
                data = b''
                if self._parse_firstline(first_line):
                    self._buf = [rest]
                else:
                    return nb_parsed
            elif not self.__on_headers_complete:
                if data:
                    self._buf.append(data)
                    data = b''

                try:
                    to_parse = b''.join(self._buf)
                    ret = self._parse_headers(to_parse)

                    if ret is False:
                        return length
                    nb_parsed = nb_parsed + (len(to_parse) - ret)
                except InvalidHeader as e:
                    self.errno = INVALID_HEADER
                    self.errstr = str(e)
                    return nb_parsed
            elif not self.__on_message_complete:
                if not self.__on_message_begin:
                    self.__on_message_begin = True

                if data:
                    self._buf.append(data)
                    data = b''

                ret = self._parse_body()
                if ret is None:
                    return length

                if ret < 0:
                    return ret
                if ret == 0:
                    self.__on_message_complete = True
                    return length
                nb_parsed = max(length, ret)

            else:
                return 0

    def _parse_firstline(self, line):
        try:
            if self.kind == 2:  # auto detect
                try:
                    self._parse_request_line(line)
                except InvalidRequestLine:
                    self._parse_response_line(line)
            elif self.kind == 1:
                self._parse_response_line(line)
            elif self.kind == 0:
                self._parse_request_line(line)
        except InvalidRequestLine as e:
            self.errno = BAD_FIRST_LINE
            self.errstr = str(e)
            return False
        return True

    def _parse_response_line(self, line):
        bits = line.split(None, 1)
        if len(bits) != 2:
            raise InvalidRequestLine(line)

        # version
        matchv = VERSION_RE.match(bits[0])
        if matchv is None:
            raise InvalidRequestLine('Invalid HTTP version: %s' % bits[0])
        self._version = (int(matchv.group(1)), int(matchv.group(2)))

        # status
        matchs = STATUS_RE.match(bits[1])
        if matchs is None:
            raise InvalidRequestLine(f'Invalid status: {bits[1]!r}')

        self._status = bits[1]
        self._status_code = int(matchs.group(1))
        self._reason = matchs.group(2)

    def _parse_request_line(self, line):
        bits = line.split(None, 2)
        if len(bits) != 3:
            raise InvalidRequestLine(line)

        # Method
        if not METHOD_RE.match(bits[0]):
            raise InvalidRequestLine('invalid Method: %s' % bits[0])
        self._method = bits[0].upper()

        # URI
        self._url = bits[1]
        parts = urlsplit(bits[1])
        self._scheme = parts.scheme or None
        self._path = parts.path or ''
        self._query_string = parts.query or ''
        if parts.fragment:
            raise InvalidRequestLine(
                'HTTP requests may not contain fragment(s)',
            )

        # Version
        match = VERSION_RE.match(bits[2])
        if match is None:
            raise InvalidRequestLine('Invalid HTTP version: %s' % bits[2])
        self._version = (int(match.group(1)), int(match.group(2)))

        # update environ
        if hasattr(self, '_environ'):
            self._environ.update(
                {
                    'PATH_INFO': self._path,
                    'QUERY_STRING': self._query_string,
                    'RAW_URI': self._url,
                    'REQUEST_METHOD': self._method,
                    'SERVER_PROTOCOL': bits[2],
                }
            )

    def _parse_headers(self, data):
        if data == b'\r\n':
            self.__on_headers_complete = True
            self._buf = []
            return 0
        idx = data.find(b'\r\n\r\n')
        if idx < 0:  # we don't have all headers
            if self._status_code == 204 and data == b'\r\n':
                self._buf = []
                self.__on_headers_complete = True
                return 0
            return False

        # Split lines on \r\n keeping the \r\n on each line
        lines = [str(line, 'unicode_escape') + '\r\n' for line in data[:idx].split(b'\r\n')]

        # Parse headers into key/value pairs paying attention
        # to continuation lines.
        while len(lines):
            # Parse initial header name : value pair.
            curr = lines.pop(0)
            if curr.find(':') < 0:
                raise InvalidHeader('invalid line %s' % curr.strip())
            name, value = curr.split(':', 1)
            name = name.rstrip(' \t').upper()
            if HEADER_RE.search(name):
                raise InvalidHeader('invalid header name %s' % name)

            if value.endswith('\r\n'):
                value = value[:-2]

            name, value = name.strip(), [value.lstrip()]

            # Consume value continuation lines
            while len(lines) and lines[0].startswith((' ', '\t')):
                curr = lines.pop(0)
                if curr.endswith('\r\n'):
                    curr = curr[:-2]
                value.append(curr)
            value = ''.join(value).rstrip()

            # store new header value
            self._headers.add_header(name, value)

            # update WSGI environ
            key = 'HTTP_%s' % name.upper().replace('-', '_')
            self._environ[key] = value

        # detect now if body is sent by chunks.
        clen = self._headers.get('content-length')
        te = self._headers.get('transfer-encoding', '').lower()

        if clen is not None:
            with contextlib.suppress(ValueError):
                self._clen_rest = self._clen = int(clen)
        else:
            self._chunked = te == 'chunked'
            if not self._chunked:
                self._clen_rest = maxsize

        # detect encoding and set decompress object
        encoding = self._headers.get('content-encoding')
        if self.decompress:
            if encoding == 'gzip':
                self.__decompress_obj = zlib.decompressobj(16 + zlib.MAX_WBITS)
                self.__decompress_first_try = False
            elif encoding == 'deflate':
                self.__decompress_obj = zlib.decompressobj()

        rest = data[idx + 4 :]
        self._buf = [rest]
        self.__on_headers_complete = True
        return len(rest)

    def _parse_body(self):
        if self._status_code == 204 and len(self._buf) == 0:
            self.__on_message_complete = True
            return None
        if not self._chunked:
            body_part = b''.join(self._buf)
            if not body_part and self._clen is None:
                if not self._status:  # message complete only for servers
                    self.__on_message_complete = True
                return None
            self._clen_rest -= len(body_part)

            # maybe decompress
            if self.__decompress_obj is not None:
                if not self.__decompress_first_try:
                    body_part = self.__decompress_obj.decompress(body_part)
                else:
                    try:
                        body_part = self.__decompress_obj.decompress(body_part)
                    except zlib.error:
                        self.__decompress_obj.decompressobj = zlib.decompressobj(-zlib.MAX_WBITS)
                        body_part = self.__decompress_obj.decompress(body_part)
                    self.__decompress_first_try = False

            self._partial_body = True
            self._body.append(body_part)
            self._buf = []

            if self._clen_rest <= 0:
                self.__on_message_complete = True
            return None
        data = b''.join(self._buf)
        try:
            size, rest = self._parse_chunk_size(data)
        except InvalidChunkSize as e:
            self.errno = INVALID_CHUNK
            self.errstr = 'invalid chunk size [%s]' % str(e)
            return -1

        if size == 0:
            return size

        if size is None or len(rest) < size:
            return None

        body_part, rest = rest[:size], rest[size:]
        if len(rest) < 2:
            self.errno = INVALID_CHUNK
            self.errstr = 'chunk missing terminator [%s]' % data
            return -1

        # maybe decompress
        if self.__decompress_obj is not None:
            body_part = self.__decompress_obj.decompress(body_part)

        self._partial_body = True
        self._body.append(body_part)

        self._buf = [rest[2:]]
        return len(rest)

    def _parse_chunk_size(self, data):
        idx = data.find(b'\r\n')
        if idx < 0:
            return None, None
        line, rest_chunk = data[:idx], data[idx + 2 :]
        chunk_size = line.split(b';', 1)[0].strip()
        try:
            chunk_size = int(chunk_size, 16)
        except ValueError:
            raise InvalidChunkSize(chunk_size)

        if chunk_size == 0:
            self._parse_trailers(rest_chunk)
            return 0, None
        return chunk_size, rest_chunk

    def _parse_trailers(self, data):
        idx = data.find(b'\r\n\r\n')

        if data[:2] == b'\r\n':
            self._trailers = self._parse_headers(data[:idx])



#!/usr/bin/env python
import pytest

from circuits import Event, Manager, handler


class foo(Event):
    """foo Event"""


@handler('foo')
def on_foo(self):
    return 'Hello World!'


def test_addHandler():
    m = Manager()
    m.start()

    m.addHandler(on_foo)

    waiter = pytest.WaitEvent(m, 'foo')
    x = m.fire(foo())
    waiter.wait()

    s = x.value
    assert s == 'Hello World!'

    m.stop()


def test_removeHandler():
    m = Manager()
    m.start()

    method = m.addHandler(on_foo)

    waiter = pytest.WaitEvent(m, 'foo')
    x = m.fire(foo())
    waiter.wait()

    s = x.value
    assert s == 'Hello World!'

    m.removeHandler(method)

    waiter = pytest.WaitEvent(m, 'foo')
    x = m.fire(foo())
    waiter.wait()

    assert x.value is None

    assert on_foo not in dir(m)
    assert 'foo' not in m._handlers

    m.stop()



"""
XML RPC

This module implements a XML RPC dispatcher that translates incoming
RPC calls over XML into RPC events.
"""

from xmlrpc.client import Fault, dumps, loads

from circuits import BaseComponent, Event, handler


class rpc(Event):
    """rpc Event"""


class XMLRPC(BaseComponent):
    channel = 'web'

    def __init__(self, path=None, encoding='utf-8', rpc_channel='*'):
        super().__init__()

        self.path = path
        self.encoding = encoding
        self.rpc_channel = rpc_channel

    @handler('request', priority=0.2)
    def _on_request(self, event, req, res):
        if self.path is not None and self.path != req.path.rstrip('/'):
            return

        res.headers['Content-Type'] = 'text/xml'

        try:
            data = req.body.read()
            params, method = loads(data)

            if not isinstance(method, str):
                method = str(method)

            value = yield self.call(rpc.create(method, *params), self.rpc_channel)
            yield self._response(value.value)
        except Exception as exc:
            yield self._error(1, f'{type(exc).__name__}: {exc}')
        finally:
            event.stop()

    def _response(self, result):
        return dumps((result,), encoding=self.encoding, allow_none=True)

    def _error(self, code, message):
        fault = Fault(code, message)
        return dumps(fault, encoding=self.encoding, allow_none=True)



#!/usr/bin/env python
from http.client import HTTPConnection

import pytest

from circuits.web import Controller

from .helpers import HTTPError, urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test_root(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_badpath_notfound(webapp):
    url = '%s/../../../../../../etc/passwd' % webapp.server.http.base
    with pytest.raises(HTTPError) as exc:
        urlopen(url)
    assert exc.value.code == 404


def test_badpath_redirect(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.connect()

    path = '/../../../../../../etc/passwd'

    connection.request('GET', path)
    response = connection.getresponse()
    assert response.status == 301
    assert response.reason == 'Moved Permanently'

    connection.close()



from urllib.parse import urlparse

from circuits.core import BaseComponent, Event, handler
from circuits.net.events import close, connect, write
from circuits.net.sockets import TCPClient
from circuits.protocols.http import HTTP
from circuits.web.headers import Headers


def parse_url(url):
    p = urlparse(url)

    if p.hostname:
        host = p.hostname
    else:
        raise ValueError('URL must be absolute')

    if p.scheme == 'http':
        secure = False
        port = p.port or 80
    elif p.scheme == 'https':
        secure = True
        port = p.port or 443
    else:
        raise ValueError('Invalid URL scheme')

    path = p.path or '/'

    if p.query:
        path += '?' + p.query

    return (host, port, path, secure)


class HTTPException(Exception):
    pass


class NotConnected(HTTPException):
    pass


class request(Event):
    """
    request Event

    This Event is used to initiate a new request.

    :param method: HTTP Method (PUT, GET, POST, DELETE)
    :type  method: str

    :param url: Request URL
    :type  url: str
    """

    def __init__(self, method, path, body=None, headers=None):
        """x.__init__(...) initializes x; see x.__class__.__doc__ for signature"""
        super().__init__(method, path, body, headers)


class Client(BaseComponent):
    channel = 'client'

    def __init__(self, channel=channel):
        super().__init__(channel=channel)
        self._response = None

        self._transport = TCPClient(channel=channel).register(self)

        HTTP(channel=channel).register(self._transport)

    @handler('write')
    def write(self, data):
        if self._transport.connected:
            self.fire(write(data), self._transport)

    @handler('close')
    def close(self):
        if self._transport.connected:
            self.fire(close(), self._transport)

    @handler('connect', priority=1)
    def connect(self, event, host=None, port=None, secure=None):
        if not self._transport.connected:
            self.fire(connect(host, port, secure), self._transport)

        event.stop()

    @handler('request')
    def request(self, method, url, body=None, headers=None):
        host, port, path, secure = parse_url(url)

        if not self._transport.connected:
            self.fire(connect(host, port, secure))
            yield self.wait('connected', self._transport.channel)

        headers = Headers(list((headers or {}).items()))

        # Clients MUST include Host header in HTTP/1.1 requests (RFC 2616)
        if 'Host' not in headers:
            headers['Host'] = '{}{}'.format(
                host,
                '' if port in (80, 443) else f':{port:d}',
            )

        if body is not None:
            headers['Content-Length'] = len(body)

        command = f'{method} {path} HTTP/1.1'
        message = f'{command}\r\n{headers}'
        self.fire(write(message.encode('utf-8')), self._transport)
        if body is not None:
            self.fire(write(body), self._transport)

        yield (yield self.wait('response'))

    @handler('response')
    def _on_response(self, response):
        self._response = response
        if response.headers.get('Connection', '').lower() == 'close':
            self.fire(close(), self._transport)
        return response

    @property
    def connected(self):
        if hasattr(self, '_transport'):
            return self._transport.connected
        return None

    @property
    def response(self):
        return getattr(self, '_response', None)



"""
Session Components

This module implements Session Components that can be used to store
and access persistent information.
"""

from abc import ABC, abstractmethod
from collections import defaultdict
from hashlib import sha1 as sha
from uuid import uuid4 as uuid

from circuits import Component, handler


def who(request, encoding='utf-8'):
    """Create a SHA1 Hash of the User's IP Address and User-Agent"""
    ip = request.remote.ip
    agent = request.headers.get('User-Agent', '')

    return sha(f'{ip}{agent}'.encode(encoding)).hexdigest()


def create_session(request):
    """
    Create a unique session id from the request

    Returns a unique session using ``uuid4()`` and a ``sha1()`` hash
    of the users IP Address and User Agent in the form of ``sid/who``.
    """
    return f'{uuid().hex}/{who(request)}'


def verify_session(request, sid):
    """
    Verify a User's Session

    This verifies the User's Session by verifying the SHA1 Hash
    of the User's IP Address and User-Agent match the provided
    Session ID.
    """
    if '/' not in sid:
        return create_session(request)

    user = sid.split('/', 1)[1]

    if user != who(request):
        return create_session(request)

    return sid


class Session(dict):
    def __init__(self, sid, data, store):
        super().__init__(data)

        self._sid = sid
        self._store = store

    @property
    def sid(self):
        return self._sid

    @property
    def store(self):
        return self._store

    def expire(self):
        self.store.delete(self.sid)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.store.save(self.sid, self)


class Store(ABC):
    @abstractmethod
    def delete(self, sid):
        """Delete the session data identified by sid"""

    @abstractmethod
    def load(self, sid):
        """Load the session data identified by sid"""

    @abstractmethod
    def save(self, sid):
        """Save the session data identified by sid"""


class MemoryStore(Store):
    def __init__(self):
        self._data = defaultdict(dict)

    @property
    def data(self):
        return self._data

    def delete(self, sid):
        del self.data[sid]

    def load(self, sid):
        return Session(sid, self.data[sid], self)

    def save(self, sid, data):
        self.data[sid] = data


class Sessions(Component):
    channel = 'web'

    def __init__(self, name='circuits', store=MemoryStore, channel=channel):
        super().__init__(channel=channel)

        self._name = name
        self._store = store()

    @property
    def name(self):
        return self._name

    @property
    def store(self):
        return self._store

    @handler('request', priority=10)
    def request(self, request, response):
        if self.name in request.cookie:
            sid = request.cookie[self._name].value
            sid = verify_session(request, sid)
        else:
            sid = create_session(request)

        request.session = self.store.load(sid)
        response.cookie[self.name] = sid



#!/usr/bin/env python
import os
import select
import sys
import tempfile

import pytest
from pytest import fixture

from circuits import Manager
from circuits.core.pollers import EPoll, KQueue, Poll, Select
from circuits.net.sockets import UNIXClient, UNIXServer, close, connect, write

from .client import Client
from .server import Server


if sys.platform in ('win32', 'cygwin'):
    pytest.skip('Test Not Applicable on Windows')


@fixture()
def tmpfile(request):
    tmpdir = tempfile.mkdtemp()
    return os.path.join(tmpdir, 'test.sock')


def pytest_generate_tests(metafunc):
    poller = [Select]

    if hasattr(select, 'poll'):
        poller.append(Poll)

    if hasattr(select, 'epoll'):
        poller.append(EPoll)

    if hasattr(select, 'kqueue'):
        poller.append(KQueue)
    metafunc.parametrize('Poller', poller)


def test_unix(tmpfile, Poller):
    m = Manager() + Poller()

    server = Server() + UNIXServer(tmpfile)
    client = Client() + UNIXClient()

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(server, 'ready')
        assert pytest.wait_for(client, 'ready')

        client.fire(connect(tmpfile))
        assert pytest.wait_for(client, 'connected')
        assert pytest.wait_for(server, 'connected')
        assert pytest.wait_for(client, 'data', b'Ready')

        client.fire(write(b'foo'))
        assert pytest.wait_for(server, 'data', b'foo')

        client.fire(close())
        assert pytest.wait_for(client, 'disconnected')
        assert pytest.wait_for(server, 'disconnected')

        server.fire(close())
        assert pytest.wait_for(server, 'closed')
    finally:
        m.stop()



"""
Utils

This module defines utilities used by circuits.
"""

import sys
from importlib import reload


def flatten(root, visited=None):
    if not visited:
        visited = set()
    yield root
    for component in root.components.copy():
        if component not in visited:
            visited.add(component)
            yield from flatten(component, visited)


def findchannel(root, channel, all=False):
    components = [x for x in flatten(root) if x.channel == channel]

    if all:
        return components

    if components:
        return components[0]
    return None


def findtype(root, component, all=False):
    components = [x for x in flatten(root) if issubclass(type(x), component)]

    if all:
        return components

    if components:
        return components[0]
    return None


findcmp = findtype


def findroot(component):
    if component.parent == component:
        return component
    return findroot(component.parent)


def safeimport(name):
    modules = sys.modules.copy()
    try:
        if name in sys.modules:
            return reload(sys.modules[name])
        return __import__(name, globals(), locals(), [''])
    except Exception:
        for name in sys.modules.copy():
            if name not in modules:
                del sys.modules[name]



#!/usr/bin/env python
from circuits.web import JSONController, Server


class Root(JSONController):
    def index(self):
        return {'success': True, 'message': 'Hello World!'}


app = Server(('0.0.0.0', 8000))
Root().register(app)
app.run()



"""Circuits events for STOMP Client"""

import logging

from circuits import Event


LOG = logging.getLogger(__name__)


class stomp_event(Event):
    """A Circuits event with less verbose repr"""

    success = True

    def _repr(self):
        return ''

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ''

        data = self._repr()

        return f'<{self.name}[{channels}] ({data})>'


class disconnected(stomp_event):
    def __init__(self, reconnect=True, receipt=None):
        super().__init__(receipt=receipt)
        self.reconnect = reconnect


class disconnect(stomp_event):
    pass


class message(stomp_event):
    def __init__(self, frame):
        super().__init__(headers=frame.headers, message=frame.body)
        self.frame = frame


class send(stomp_event):
    def __init__(self, headers, body, destination):
        super().__init__(headers=headers, body=body, destination=destination)


class client_heartbeat(stomp_event):
    pass


class server_heartbeat(stomp_event):
    pass


class connect(stomp_event):
    def __init__(self, subscribe=False, host=None):
        super().__init__(host=host)
        self.subscribe = subscribe


class connected(stomp_event):
    pass


class connection_failed(stomp_event):
    reconnect = True


class on_stomp_error(stomp_event):
    def __init__(self, frame, err):
        headers = frame.headers if frame else {}
        body = frame.body if frame else None
        super().__init__(headers=headers, message=body, error=err)
        self.frame = frame


class heartbeat_timeout(stomp_event):
    pass


class subscribe(stomp_event):
    def __init__(self, destination, **kwargs):
        super().__init__(destination=destination, **kwargs)
        self.destination = destination


class unsubscribe(stomp_event):
    def __init__(self, destination):
        super().__init__(destination=destination)
        self.destination = destination


class ack(stomp_event):
    def __init__(self, frame):
        super().__init__(frame=frame)
        self.frame = frame



"""Documentation Tasks"""

from fabric.api import lcd, local, task

from .utils import pip, requires


PACKAGE = 'circuits'


@task()
def api():
    """Generate the API Documentation"""
    if PACKAGE is not None:
        pip(requirements='docs/requirements.txt')
        local(f'sphinx-apidoc -f -e -T -o docs/source/api {PACKAGE:s}')


@task()
@requires('make')
def clean():
    """Delete Generated Documentation"""
    with lcd('docs'):
        pip(requirements='requirements.txt')
        local('make clean')


@task(default=True)
@requires('make')
def build(**options):
    """Build the Documentation"""
    pip(requirements='docs/requirements.txt')

    with lcd('docs'):
        local('make html')


@task()
@requires('open')
def view(**options):
    """View the Documentation"""
    with lcd('docs'):
        import webbrowser

        webbrowser.open_new_tab('build/html/index.html')



"""py.test config"""

import os

import pytest

from circuits import Component, Debugger, handler
from circuits.net.sockets import close
from circuits.web import Server, Static
from circuits.web.client import Client, request


DOCROOT = os.path.join(os.path.dirname(__file__), 'static')


class WebApp(Component):
    channel = 'web'

    def init(self):
        self.closed = False

        self.server = Server(0).register(self)
        Static('/static', DOCROOT, dirlisting=True).register(self)


class WebClient(Client):
    def init(self, *args, **kwargs):
        self.closed = False

    def __call__(self, method, path, body=None, headers={}):
        waiter = pytest.WaitEvent(self, 'response', channel=self.channel)
        self.fire(request(method, path, body, headers))
        assert waiter.wait()

        return self.response

    @handler('closed', channel='*', priority=1.0)
    def _on_closed(self):
        self.closed = True


@pytest.fixture()
def webapp(request, manager, watcher):
    webapp = WebApp().register(manager)
    assert watcher.wait('ready')

    if hasattr(request.module, 'application'):
        from circuits.web.wsgi import Gateway

        application = request.module.application
        Gateway({'/': application}).register(webapp)
        assert watcher.wait('registered')

    Root = getattr(request.module, 'Root', None)
    if Root is not None:
        Root().register(webapp)
        assert watcher.wait('registered')

    if request.config.option.verbose:
        Debugger().register(webapp)
        assert watcher.wait('registered')

    def finalizer():
        webapp.fire(close())
        assert watcher.wait('closed')

        webapp.unregister()
        assert watcher.wait('unregistered')

    request.addfinalizer(finalizer)

    return webapp



#!/usr/bin/env python
"""
Multi Bridge Example

Identical to the Hello Bridge Example but with a 2nd child.
"""

from os import getpid

from circuits import Component, Event, ipc


class go(Event):
    """go Event"""


class hello(Event):
    """hello Event"""


class Child(Component):
    def hello(self):
        return f'Hello from child with pid {getpid()}'


class App(Component):
    def init(self):
        self.counter = 0
        self.child1 = Child().start(process=True, link=self)
        self.child2 = Child().start(process=True, link=self)

    def ready(self, *args):
        self.counter += 1
        if self.counter < 2:
            return
        self.fire(go())

    def go(self):
        x = yield self.call(hello())
        yield print(x)

        y = yield self.call(ipc(hello()), self.child1[1].channel)
        yield print(y)

        z = yield self.call(ipc(hello()), self.child2[1].channel)
        yield print(z)

        raise SystemExit(0)

    def hello(self):
        return f'Hello from parent with pid {getpid()}'


App().run()



"""
Socket Components

This module contains various Socket Components for use with Networking.
"""

import contextlib
import os
import select
from collections import defaultdict, deque
from errno import (
    EAGAIN,
    EALREADY,
    EBADF,
    ECONNABORTED,
    EINPROGRESS,
    EINTR,
    EINVAL,
    EISCONN,
    EMFILE,
    ENFILE,
    ENOBUFS,
    ENOMEM,
    ENOTCONN,
    EPERM,
    EPIPE,
    EWOULDBLOCK,
)
from socket import (
    AF_INET,
    AF_INET6,
    IPPROTO_IP,
    IPPROTO_TCP,
    SO_BROADCAST,
    SO_REUSEADDR,
    SOCK_DGRAM,
    SOCK_STREAM,
    SOL_SOCKET,
    TCP_NODELAY,
    gaierror,
    getaddrinfo,
    getfqdn,
    gethostbyname,
    gethostname,
    socket,
)
from time import time

from _socket import socket as SocketType

from circuits.core import BaseComponent, handler
from circuits.core.pollers import BasePoller, Poller
from circuits.core.utils import findcmp

from .events import close, closed, connect, connected, disconnect, disconnected, error, read, ready, unreachable, write


try:
    from socket import AF_UNIX
except ImportError:
    AF_UNIX = None

try:
    from socket import socketpair
except ImportError:
    socketpair = None

try:
    from ssl import (
        CERT_NONE,
        CERT_REQUIRED,
        PROTOCOL_TLS,
        SSL_ERROR_WANT_READ,
        SSL_ERROR_WANT_WRITE,
        Purpose,
        SSLContext,
        SSLError,
        create_default_context,
    )

    HAS_SSL = 1
except ImportError:
    import warnings

    warnings.warn('No SSL support available.')
    HAS_SSL = 0
    CERT_NONE = None
    PROTOCOL_TLS = None


BUFSIZE = 4096  # 4KB Buffer
BACKLOG = 5000  # 5K Concurrent Connections


def wrap_socket(
    sock,
    keyfile=None,
    certfile=None,
    ca_certs=None,
    purpose=Purpose.SERVER_AUTH,
    verify_mode=None,
    minimum_version=PROTOCOL_TLS,
    check_hostname=None,
    **kwargs,
):
    if minimum_version != PROTOCOL_TLS:
        context = SSLContext(minimum_version)
        if purpose == Purpose.SERVER_AUTH:
            context.verify_mode = CERT_REQUIRED
            context.check_hostname = True
        if ca_certs:
            context.load_verify_locations(cafile=ca_certs)
    else:
        context = create_default_context(purpose, cafile=ca_certs)
    if check_hostname is not None:
        context.check_hostname = check_hostname
    if verify_mode:
        if verify_mode == CERT_NONE:
            context.check_hostname = False
        context.verify_mode = verify_mode
    if certfile:
        context.load_cert_chain(certfile=certfile, keyfile=keyfile)
    return context.wrap_socket(sock, **kwargs)


def do_handshake(sock, on_done=None, on_error=None, extra_args=None):
    """
    SSL Async Handshake

    :param on_done: Function called when handshake is complete
    :type on_done: :function:

    :param on_error: Function called when handshake errored
    :type on_error: :function:
    """
    extra_args = extra_args or ()

    while True:
        try:
            sock.do_handshake()
            break
        except SSLError as err:
            if err.args[0] == SSL_ERROR_WANT_READ:
                select.select([sock], [], [])
            elif err.args[0] == SSL_ERROR_WANT_WRITE:
                select.select([], [sock], [])
            else:
                callable(on_error) and on_error(sock, err)
                return

        yield

    callable(on_done) and on_done(sock, *extra_args)


class Client(BaseComponent):
    channel = 'client'

    socket_family = AF_INET
    socket_type = SOCK_STREAM
    socket_protocol = IPPROTO_IP
    socket_options = []

    def __init__(self, bind=None, bufsize=BUFSIZE, channel=channel, **kwargs):
        super().__init__(channel=channel, **kwargs)

        if isinstance(bind, SocketType):
            self._bind = bind.getsockname()
            self._sock = bind
        else:
            self._bind = self.parse_bind_parameter(bind)
            self._sock = self._create_socket()

        self._bufsize = bufsize

        self._ssock = None
        self._poller = None
        self._buffer = deque()
        self._closeflag = False
        self._connected = False

        self.host = None
        self.port = 0
        self.secure = False

        self.server = {}
        self.issuer = {}

    def parse_bind_parameter(self, bind_parameter):
        return parse_ipv4_parameter(bind_parameter)

    @property
    def connected(self):
        return getattr(self, '_connected', None)

    @handler('registered', 'started', channel='*')
    def _on_registered_or_started(self, component, manager=None):
        if self._poller is None:
            if isinstance(component, BasePoller):
                self._poller = component
                self.fire(ready(self))
            else:
                if component is not self:
                    return
                component = findcmp(self.root, BasePoller)
                if component is not None:
                    self._poller = component
                    self.fire(ready(self))
                else:
                    self._poller = Poller().register(self)
                    self.fire(ready(self))

    @handler('stopped', channel='*')
    def _on_stopped(self, component):
        self.fire(close())

    @handler('read_value_changed')
    def _on_read_value_changed(self, value):
        if isinstance(value, bytes):
            self.fire(write(value))

    @handler('prepare_unregister', channel='*')
    def _on_prepare_unregister(self, event, c):
        if event.in_subtree(self):
            self._close()

    def _close(self):
        if not self._connected:
            return

        self._poller.discard(self._sock)

        self._buffer.clear()
        self._closeflag = False
        self._connected = False

        with contextlib.suppress(OSError):
            self._sock.shutdown(2)
        with contextlib.suppress(OSError):
            self._sock.close()

        self.fire(disconnected())

    @handler('close')
    def close(self):
        if not self._buffer:
            self._close()
        elif not self._closeflag:
            self._closeflag = True

    def _read(self):
        try:
            try:
                data = self._ssock.read(self._bufsize) if self.secure and self._ssock else self._sock.recv(self._bufsize)
            except SSLError as exc:
                if exc.errno in (SSL_ERROR_WANT_READ, SSL_ERROR_WANT_WRITE):
                    return
                raise

            if data:
                self.fire(read(data)).notify = True
            else:
                self.close()
        except OSError as e:
            if e.args[0] == EWOULDBLOCK:
                return
            self.fire(error(e))
            self._close()

    def _write(self, data):
        try:
            nbytes = self._ssock.write(data) if self.secure and self._ssock else self._sock.send(data)

            if nbytes < len(data):
                self._buffer.appendleft(data[nbytes:])
        except OSError as e:
            if e.args[0] in (EPIPE, ENOTCONN):
                self._close()
            else:
                self.fire(error(e))

    @handler('write')
    def write(self, data):
        if not self._poller.isWriting(self._sock):
            self._poller.addWriter(self, self._sock)
        self._buffer.append(data)

    @handler('_disconnect', priority=1)
    def __on_disconnect(self, sock):
        self._close()

    @handler('_read', priority=1)
    def __on_read(self, sock):
        self._read()

    @handler('_write', priority=1)
    def __on_write(self, sock):
        if self._buffer:
            data = self._buffer.popleft()
            self._write(data)

        if not self._buffer:
            if self._closeflag:
                self._close()
            elif self._poller.isWriting(self._sock):
                self._poller.removeWriter(self._sock)

    def _create_socket(self):
        sock = socket(self.socket_family, self.socket_type, self.socket_protocol)

        for option in self.socket_options:
            sock.setsockopt(*option)
        sock.setblocking(False)
        if self._bind is not None:
            sock.bind(self._bind)
        return sock


class TCPClient(Client):
    socket_family = AF_INET
    socket_type = SOCK_STREAM
    socket_protocol = IPPROTO_TCP
    socket_options = [
        (IPPROTO_TCP, TCP_NODELAY, 1),
    ]

    def init(self, connect_timeout=5, *args, **kwargs):
        self.connect_timeout = connect_timeout

    @handler('connect')  # noqa
    def connect(self, host, port, secure=False, **kwargs):
        # TODO: C901: This has a high McCacbe complexity score of 10.
        # TODO: Refactor this!

        self.host = host
        self.port = port
        self.secure = secure

        if self.secure:
            self.certfile = kwargs.get('certfile', None)
            self.keyfile = kwargs.get('keyfile', None)
            self.ca_certs = kwargs.get('ca_certs', None)

        try:
            r = self._sock.connect((host, port))
        except OSError as e:
            if e.args[0] in (EBADF, EINVAL):
                self._sock = self._create_socket()
                r = self._sock.connect_ex((host, port))
            else:
                r = e.args[0]

            if r not in (EISCONN, EWOULDBLOCK, EINPROGRESS, EALREADY):
                self.fire(unreachable(host, port, e))
                self.fire(error(e))
                self._close()
                return

        stop_time = time() + self.connect_timeout
        while time() < stop_time:
            try:
                self._sock.getpeername()
                self._connected = True
                break
            except Exception:
                yield

        if not self._connected:
            self.fire(unreachable(host, port))
            return

        def on_done(sock):
            self._poller.addReader(self, sock)
            self.fire(connected(host, port))

        if self.secure:

            def on_error(sock, err):
                self.fire(error(sock, err))
                self._close()

            self._sock = wrap_socket(
                self._sock,
                self.keyfile,
                self.certfile,
                ca_certs=self.ca_certs,
                purpose=Purpose.SERVER_AUTH,
                server_hostname=self.host,
                do_handshake_on_connect=False,
            )
            for _ in do_handshake(self._sock, on_done, on_error):
                yield
        else:
            on_done(self._sock)


class TCP6Client(TCPClient):
    socket_family = AF_INET6

    def parse_bind_parameter(self, bind_parameter):
        return parse_ipv6_parameter(bind_parameter)


class UNIXClient(Client):
    socket_family = AF_UNIX
    socket_type = SOCK_STREAM
    socket_options = []

    @handler('ready')
    def ready(self, component):
        if self._poller is not None and self._connected:
            self._poller.addReader(self, self._sock)

    @handler('connect')  # noqa
    def connect(self, path, secure=False, **kwargs):
        # TODO: C901: This has a high McCacbe complexity score of 10.
        # TODO: Refactor this!

        self.path = path
        self.secure = secure

        if self.secure:
            self.certfile = kwargs.get('certfile', None)
            self.keyfile = kwargs.get('keyfile', None)
            self.ca_certs = kwargs.get('ca_certs', None)

        try:
            r = self._sock.connect_ex(path)
        except OSError as e:
            r = e.args[0]

        if r:
            if r in (EISCONN, EWOULDBLOCK, EINPROGRESS, EALREADY):
                self._connected = True
            else:
                self.fire(error(r))
                return

        self._connected = True

        self._poller.addReader(self, self._sock)

        if self.secure:

            def on_done(sock):
                self.fire(connected(gethostname(), path))

            def on_error(sock, err):
                self.fire(error(err))

            self._ssock = wrap_socket(
                self._sock,
                self.keyfile,
                self.certfile,
                ca_certs=self.ca_certs,
                purpose=Purpose.SERVER_AUTH,
                server_hostname=self.host,
                do_handshake_on_connect=False,
            )
            for _ in do_handshake(self._ssock, on_done, on_error):
                yield
        else:
            self.fire(connected(gethostname(), path))


class Server(BaseComponent):
    channel = 'server'
    socket_protocol = IPPROTO_IP

    def __init__(self, bind, secure=False, backlog=BACKLOG, bufsize=BUFSIZE, channel=channel, **kwargs):
        super().__init__(channel=channel)

        self.socket_options = self.socket_options[:] + kwargs.get('socket_options', [])
        self._bind = self.parse_bind_parameter(bind)

        self._backlog = backlog
        self._bufsize = bufsize

        if isinstance(bind, socket):
            self._sock = bind
        else:
            self._sock = self._create_socket()

        self._closeq = []
        self._clients = []
        self._poller = None
        self._buffers = defaultdict(deque)

        self.__starttls = set()

        self.secure = secure
        self.certfile = kwargs.get('certfile')
        self.keyfile = kwargs.get('keyfile', None)
        self.cert_reqs = kwargs.get('cert_reqs', CERT_NONE)
        self.ssl_version = kwargs.get('ssl_version', PROTOCOL_TLS)
        self.ca_certs = kwargs.get('ca_certs', None)
        if self.secure and not self.certfile:
            raise RuntimeError('certfile must be specified for server-side operations')

    def parse_bind_parameter(self, bind_parameter):
        return parse_ipv4_parameter(bind_parameter)

    @property
    def connected(self):
        return True

    @property
    def host(self):
        if getattr(self, '_sock', None) is not None:
            try:
                sockname = self._sock.getsockname()
                if isinstance(sockname, tuple):
                    return sockname[0]
                return sockname
            except OSError:
                return None
        return None

    @property
    def port(self):
        if getattr(self, '_sock', None) is not None:
            try:
                sockname = self._sock.getsockname()
                if isinstance(sockname, tuple):
                    return sockname[1]
            except OSError:
                return None
        return None

    @handler('registered', 'started', channel='*')
    def _on_registered_or_started(self, component, manager=None):
        if self._poller is None:
            if isinstance(component, BasePoller):
                self._poller = component
                self._poller.addReader(self, self._sock)
                self.fire(ready(self, (self.host, self.port)))
            else:
                if component is not self:
                    return
                component = findcmp(self.root, BasePoller)
                if component is not None:
                    self._poller = component
                    self._poller.addReader(self, self._sock)
                    self.fire(ready(self, (self.host, self.port)))
                else:
                    try:
                        self._poller = Poller().register(self)
                    except OSError as err:
                        self.fire(error(err))
                    else:
                        self._poller.addReader(self, self._sock)
                        self.fire(ready(self, (self.host, self.port)))

    @handler('stopped', channel='*')
    def _on_stopped(self, component):
        self.fire(close())

    @handler('read_value_changed')
    def _on_read_value_changed(self, value):
        if isinstance(value.value, bytes):
            sock = value.event.args[0]
            self.fire(write(sock, value.value))

    def _close(self, sock):
        if sock is None:
            return

        if sock != self._sock and sock not in self._clients:
            return

        self._poller.discard(sock)

        if sock in self._buffers:
            del self._buffers[sock]

        if sock in self._clients:
            self._clients.remove(sock)
        else:
            self._sock = None

        if sock in self.__starttls:
            self.__starttls.remove(sock)

        with contextlib.suppress(OSError):
            sock.shutdown(2)
        with contextlib.suppress(OSError):
            sock.close()

        self.fire(disconnect(sock))

    @handler('close')
    def close(self, sock=None):
        is_closed = sock is None

        if sock is None:
            socks = [self._sock]
            socks.extend(self._clients[:])
        else:
            socks = [sock]

        for sock in socks:
            if not self._buffers[sock]:
                self._close(sock)
            elif sock not in self._closeq:
                self._closeq.append(sock)

        if is_closed:
            self.fire(closed())

    def _read(self, sock):
        if sock not in self._clients:
            return

        try:
            data = sock.recv(self._bufsize)
            if data:
                self.fire(read(sock, data)).notify = True
            else:
                self.close(sock)
        except OSError as e:
            if e.args[0] == EWOULDBLOCK:
                return
            self.fire(error(sock, e))
            self._close(sock)

    def _write(self, sock, data):
        if sock not in self._clients:
            return

        try:
            nbytes = sock.send(data)
            if nbytes < len(data):
                self._buffers[sock].appendleft(data[nbytes:])
        except OSError as e:
            if e.args[0] not in (EINTR, EWOULDBLOCK, ENOBUFS):
                self.fire(error(sock, e))
                self._close(sock)
            else:
                self._buffers[sock].appendleft(data)

    @handler('write')
    def write(self, sock, data):
        if not self._poller.isWriting(sock):
            self._poller.addWriter(self, sock)
        self._buffers[sock].append(data)

    def _accept(self):
        try:
            newsock, _host = self._sock.accept()
        except OSError as e:
            if e.args[0] in (EWOULDBLOCK, EAGAIN):
                return
            elif e.args[0] == EPERM:
                # Netfilter on Linux may have rejected the
                # connection, but we get told to try to accept()
                # anyway.
                return
            elif e.args[0] in (EMFILE, ENOBUFS, ENFILE, ENOMEM, ECONNABORTED):
                # Linux gives EMFILE when a process is not allowed
                # to allocate any more file descriptors.  *BSD and
                # Win32 give (WSA)ENOBUFS.  Linux can also give
                # ENFILE if the system is out of inodes, or ENOMEM
                # if there is insufficient memory to allocate a new
                # dentry.  ECONNABORTED is documented as possible on
                # both Linux and Windows, but it is not clear
                # whether there are actually any circumstances under
                # which it can happen (one might expect it to be
                # possible if a client sends a FIN or RST after the
                # server sends a SYN|ACK but before application code
                # calls accept(2), however at least on Linux this
                # _seems_ to be short-circuited by syncookies.
                return
            else:
                raise

        if self.secure and HAS_SSL:
            for _ in self._do_handshake(newsock):
                yield
        else:
            self._on_accept_done(newsock)

    def _do_handshake(self, sock, fire_connect_event=True):
        sslsock = wrap_socket(
            sock,
            self.keyfile,
            self.certfile,
            ca_certs=self.ca_certs,
            purpose=Purpose.CLIENT_AUTH,
            verify_mode=self.cert_reqs,
            minimum_version=self.ssl_version,
            server_side=True,
            do_handshake_on_connect=False,
        )

        yield from do_handshake(sslsock, self._on_accept_done, self._on_handshake_error, (fire_connect_event,))

    def _on_accept_done(self, sock, fire_connect_event=True):
        sock.setblocking(False)
        self._poller.addReader(self, sock)
        self._clients.append(sock)
        if fire_connect_event:
            try:
                self.fire(connect(sock, *sock.getpeername()))
            except OSError as exc:
                # errno 107 (ENOTCONN): the client already disconnected
                self._on_handshake_error(sock, exc)

    def _on_handshake_error(self, sock, err):
        self.fire(error(sock, err))
        self._close(sock)

    @handler('starttls')
    def starttls(self, sock):
        if not HAS_SSL:
            raise RuntimeError('Cannot start TLS. No TLS support.')
        if sock in self.__starttls:
            raise RuntimeError('Cannot reuse socket for already started STARTTLS.')
        self.__starttls.add(sock)
        self._poller.removeReader(sock)
        self._clients.remove(sock)
        for _ in self._do_handshake(sock, False):
            yield

    @handler('_disconnect', priority=1)
    def _on_disconnect(self, sock):
        self._close(sock)

    @handler('_read', priority=1)
    def _on_read(self, sock):
        if sock == self._sock:
            return self._accept()
        self._read(sock)
        return None

    @handler('_write', priority=1)
    def _on_write(self, sock):
        if self._buffers[sock]:
            data = self._buffers[sock].popleft()
            self._write(sock, data)

        if not self._buffers[sock]:
            if sock in self._closeq:
                self._closeq.remove(sock)
                self._close(sock)
            elif self._poller.isWriting(sock):
                self._poller.removeWriter(sock)

    def _create_socket(self):
        sock = socket(self.socket_family, self.socket_type, self.socket_protocol)

        for option in self.socket_options:
            sock.setsockopt(*option)
        sock.setblocking(False)
        if self._bind is not None:
            sock.bind(self._bind)
        return sock


class TCPServer(Server):
    socket_family = AF_INET
    socket_type = SOCK_STREAM
    socket_options = [
        (SOL_SOCKET, SO_REUSEADDR, 1),
        (IPPROTO_TCP, TCP_NODELAY, 1),
    ]

    def _create_socket(self):
        sock = super()._create_socket()
        sock.listen(self._backlog)

        return sock

    def parse_bind_parameter(self, bind_parameter):
        return parse_ipv4_parameter(bind_parameter)


def parse_ipv4_parameter(bind_parameter):
    if isinstance(bind_parameter, int):
        try:
            bind = (gethostbyname(gethostname()), bind_parameter)
        except gaierror:
            bind = ('0.0.0.0', bind_parameter)
    elif isinstance(bind_parameter, str) and ':' in bind_parameter:
        host, port = bind_parameter.split(':')
        port = int(port)
        bind = (host, port)
    else:
        bind = bind_parameter

    return bind


def parse_ipv6_parameter(bind_parameter):
    if isinstance(bind_parameter, int):
        try:
            _, _, _, _, bind = getaddrinfo(getfqdn(), bind_parameter, AF_INET6)[0]
        except (gaierror, IndexError):
            bind = ('::', bind_parameter)
    else:
        bind = bind_parameter

    return bind


class TCP6Server(TCPServer):
    socket_family = AF_INET6

    def parse_bind_parameter(self, bind_parameter):
        return parse_ipv6_parameter(bind_parameter)


class UNIXServer(Server):
    socket_family = AF_UNIX
    socket_type = SOCK_STREAM
    socket_options = [
        (SOL_SOCKET, SO_REUSEADDR, 1),
    ]

    def _create_socket(self):
        if os.path.exists(self._bind):
            os.unlink(self._bind)

        sock = super()._create_socket()
        sock.listen(self._backlog)

        return sock


class UDPServer(Server):
    socket_family = AF_INET
    socket_type = SOCK_DGRAM
    socket_options = [
        (SOL_SOCKET, SO_BROADCAST, 1),
        (SOL_SOCKET, SO_REUSEADDR, 1),
    ]

    def _close(self, sock):
        self._poller.discard(sock)

        if sock in self._buffers:
            del self._buffers[sock]

        with contextlib.suppress(OSError):
            sock.shutdown(2)
        with contextlib.suppress(OSError):
            sock.close()

        self.fire(disconnect(sock))

    @handler('close', override=True)
    def close(self):
        self.fire(closed())

        if self._buffers[self._sock] and self._sock not in self._closeq:
            self._closeq.append(self._sock)
        else:
            self._close(self._sock)

    def _read(self):
        try:
            data, address = self._sock.recvfrom(self._bufsize)
            if data:
                self.fire(read(address, data)).notify = True
        except OSError as e:
            if e.args[0] in (EWOULDBLOCK, EAGAIN):
                return
            self.fire(error(self._sock, e))
            self._close(self._sock)

    def _write(self, address, data):
        try:
            bytes = self._sock.sendto(data, address)
            if bytes < len(data):
                self._buffers[self._sock].appendleft(data[bytes:])
        except OSError as e:
            if e.args[0] in (EPIPE, ENOTCONN):
                self._close(self._sock)
            else:
                self.fire(error(self._sock, e))

    @handler('write', override=True)
    def write(self, address, data):
        if not self._poller.isWriting(self._sock):
            self._poller.addWriter(self, self._sock)
        self._buffers[self._sock].append((address, data))

    @handler('broadcast', override=True)
    def broadcast(self, data, port):
        self.write(('<broadcast>', port), data)

    @handler('_disconnect', priority=1, override=True)
    def _on_disconnect(self, sock):
        self._close(sock)

    @handler('_read', priority=1, override=True)
    def _on_read(self, sock):
        self._read()

    @handler('_write', priority=1, override=True)
    def _on_write(self, sock):
        if self._buffers[self._sock]:
            address, data = self._buffers[self._sock].popleft()
            self._write(address, data)

        if not self._buffers[self._sock]:
            if self._sock in self._closeq:
                self._closeq.remove(self._sock)
                self._close(self._sock)
            elif self._poller.isWriting(self._sock):
                self._poller.removeWriter(self._sock)


UDPClient = UDPServer


class UDP6Server(UDPServer):
    socket_family = AF_INET6

    def parse_bind_parameter(self, bind_parameter):
        return parse_ipv6_parameter(bind_parameter)


UDP6Client = UDP6Server


def Pipe(*channels, **kwargs):
    """
    Create a new full duplex Pipe

    Returns a pair of UNIXClient instances connected on either side of
    the pipe.
    """
    if socketpair is None:
        raise RuntimeError('No socketpair support available.')

    if not channels:
        channels = ('a', 'b')

    s1, s2 = socketpair()
    s1.setblocking(False)
    s2.setblocking(False)

    a = UNIXClient(s1, channel=channels[0], **kwargs)
    b = UNIXClient(s2, channel=channels[1], **kwargs)

    a._connected = True
    b._connected = True

    return a, b



#!/usr/bin/env python
from io import BytesIO

from circuits.web.utils import compress, get_ranges


try:
    from gzip import decompress
except ImportError:
    import zlib

    decompress = zlib.decompressobj(16 + zlib.MAX_WBITS).decompress  # NOQA


def test_ranges():
    assert get_ranges('bytes=3-6', 8) == [(3, 7)]
    assert get_ranges('bytes=2-4,-1', 8) == [(2, 5), (7, 8)]


def test_gzip():
    s = b'Hello World!'
    contents = BytesIO(s)
    compressed = b''.join(compress(contents, 1))
    uncompressed = decompress(compressed)
    assert uncompressed == s
    contents.close()



#!/usr/bin/env python
import sys
from subprocess import PIPE, Popen

from . import exitcodeapp


def test_ints(tmpdir):
    for expected_status in range(4):
        args = [sys.executable, exitcodeapp.__file__, f'{expected_status:d}']
        p = Popen(args, env={'PYTHONPATH': ':'.join(sys.path)})
        status = p.wait()

        assert status == expected_status


def test_string(tmpdir):
    args = [sys.executable, exitcodeapp.__file__, 'foobar']
    p = Popen(args, env={'PYTHONPATH': ':'.join(sys.path)}, stderr=PIPE)
    status = p.wait()

    assert status == 1
    assert p.stderr.read() == b'foobar\n'



"""
Debugger component used to debug each event in a system by printing
each event to sys.stderr or to a Logger Component instance.
"""

import os
import sys
from signal import SIGINT, SIGTERM
from traceback import format_exc, format_exception_only

from .components import BaseComponent
from .handlers import handler, reprhandler


class Debugger(BaseComponent):
    """
    Create a new Debugger Component

    Creates a new Debugger Component that listens to all events in the system
    printing each event to sys.stderr or a Logger Component.

    :var IgnoreEvents: list of events (str) to ignore
    :var IgnoreChannels: list of channels (str) to ignore
    :var enabled: Enabled/Disabled flag

    :param log: Logger Component instance or None (*default*)
    """

    IgnoreEvents = ['generate_events']
    IgnoreChannels = []

    def __init__(self, errors=True, events=True, file=None, logger=None, prefix=None, trim=None, **kwargs):
        """Initializes x; see x.__class__.__doc__ for signature"""
        super().__init__()

        self._errors = errors
        self._events = events

        if isinstance(file, str):
            self.file = open(os.path.abspath(os.path.expanduser(file)), 'a')
        elif hasattr(file, 'write'):
            self.file = file
        else:
            self.file = sys.stderr

        self.logger = logger
        self.prefix = prefix
        self.trim = trim

        self.IgnoreEvents.extend(kwargs.get('IgnoreEvents', []))
        self.IgnoreChannels.extend(kwargs.get('IgnoreChannels', []))

    @handler('signal', channel='*')
    def _on_signal(self, signo, stack):
        if signo in [SIGINT, SIGTERM]:
            raise SystemExit(0)

    @handler('exception', channel='*', priority=100.0)
    def _on_exception(self, error_type, value, traceback, handler=None, fevent=None):
        if not self._errors:
            return

        s = []

        handler = '' if handler is None else reprhandler(handler)

        msg = f'ERROR {handler} ({fevent!r}) ({error_type!r}): {value!r}\n'

        s.append(msg)
        s.append('Traceback (most recent call last):\n')
        s.extend(traceback)
        s.extend(format_exception_only(error_type, value))
        s.append('\n')

        if self.logger is not None:
            self.logger.error(''.join(s))
        else:
            try:
                self.file.write(''.join(s))
                self.file.flush()
            except OSError:
                pass

    @handler(priority=101.0)
    def _on_event(self, event, *args, **kwargs):
        """
        Global Event Handler

        Event handler to listen to all events printing
        each event to self.file or a Logger Component instance
        by calling self.logger.debug
        """
        try:
            if not self._events:
                return

            channels = event.channels

            if event.name in self.IgnoreEvents:
                return

            if all(channel in self.IgnoreChannels for channel in channels):
                return

            s = repr(event)

            if self.prefix:
                s = f'{self.prefix()}: {s}' if hasattr(self.prefix, '__call__') else f'{self.prefix}: {s}'

            if self.trim:
                s = '%s ...>' % s[: self.trim]

            if self.logger is not None:
                self.logger.debug(s)
            else:
                self.file.write(s)
                self.file.write('\n')
                self.file.flush()
        except Exception as e:
            sys.stderr.write(f'ERROR (Debugger): {e}')
            sys.stderr.write(f'{format_exc()}')



#!/usr/bin/env python
# curl -i http://localhost:8000/
# curl -i http://localhost:8000/ -H 'Host: foo.localdomain:8000'
# curl -i http://localhost:8000/ -H 'Host: bar.localdomain:8000'

from circuits.web import Controller, Server
from circuits.web.dispatchers import VirtualHosts


class Root(Controller):
    def index(self):
        return 'I am the main vhost'


class Foo(Controller):
    channel = '/foo'

    def index(self):
        return 'I am foo.'


class Bar(Controller):
    channel = '/bar'

    def index(self):
        return 'I am bar.'


domains = {
    'foo.localdomain:8000': 'foo',
    'bar.localdomain:8000': 'bar',
}


app = Server(('0.0.0.0', 8000))
VirtualHosts(domains).register(app)
Root().register(app)
Foo().register(app)
Bar().register(app)
app.run()



#!/usr/bin/env python
from circuits import Component
from circuits.web import JSONRPC, Controller

from .helpers import urlopen
from .jsonrpclib import ServerProxy


class App(Component):
    def eval(self, s):
        return eval(s)


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test(webapp):
    rpc = JSONRPC('/rpc')
    test = App()
    rpc.register(webapp)
    test.register(webapp)

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    url = '%s/rpc' % webapp.server.http.base
    jsonrpc = ServerProxy(url, allow_none=True, encoding='utf-8')

    data = jsonrpc.eval('1 + 2')
    assert data['result'] == 3

    rpc.unregister()
    test.unregister()



"""Internet Relay Chat Protocol numerics"""

RPL_WELCOME = 1
RPL_YOURHOST = 2

RPL_TRACELINK = 200
RPL_TRACECONNECTING = 201
RPL_TRACEHANDSHAKE = 202
RPL_TRACEUNKNOWN = 203
RPL_TRACEOPERATOR = 204
RPL_TRACEUSER = 205
RPL_TRACESERVER = 206
RPL_TRACENEWTYPE = 208
RPL_TRACELOG = 261
RPL_STATSLINKINFO = 211
RPL_STATSCOMMANDS = 212
RPL_STATSCLINE = 213
RPL_STATSNLINE = 214
RPL_STATSILINE = 215
RPL_STATSKLINE = 216
RPL_STATSYLINE = 218
RPL_ENDOFSTATS = 219
RPL_STATSLLINE = 241
RPL_STATSUPTIME = 242
RPL_STATSOLINE = 243
RPL_STATSHLINE = 244
RPL_UMODEIS = 221
RPL_LUSERCLIENT = 251
RPL_LUSEROP = 252
RPL_LUSERUNKNOWN = 253
RPL_LUSERCHANNELS = 254
RPL_LUSERME = 255
RPL_ADMINME = 256
RPL_ADMINLOC1 = 257
RPL_ADMINLOC2 = 258
RPL_ADMINEMAIL = 259

RPL_NONE = 300
RPL_USERHOST = 302
RPL_ISON = 303
RPL_AWAY = 301
RPL_UNAWAY = 305
RPL_NOWAWAY = 306
RPL_WHOISUSER = 311
RPL_WHOISSERVER = 312
RPL_WHOISOPERATOR = 313
RPL_WHOISIDLE = 317
RPL_ENDOFWHOIS = 318
RPL_WHOISCHANNELS = 319
RPL_WHOWASUSER = 314
RPL_ENDOFWHOWAS = 369
RPL_LIST = 322
RPL_LISTEND = 323
RPL_CHANNELMODEIS = 324
RPL_NOTOPIC = 331
RPL_TOPIC = 332
RPL_INVITING = 341
RPL_SUMMONING = 342
RPL_VERSION = 351
RPL_WHOREPLY = 352
RPL_ENDOFWHO = 315
RPL_NAMEREPLY = 353
RPL_ENDOFNAMES = 366
RPL_LINKS = 364
RPL_ENDOFLINKS = 365
RPL_BANLIST = 367
RPL_ENDOFBANLIST = 368
RPL_INFO = 371
RPL_ENDOFINFO = 374
RPL_MOTDSTART = 375
RPL_MOTD = 372
RPL_ENDOFMOTD = 376
RPL_YOUREOPER = 381
RPL_REHASHING = 382
RPL_TIME = 391
RPL_USERSSTART = 392
RPL_USERS = 393
RPL_ENDOFUSERS = 394
RPL_NOUSERS = 395

ERR_NOSUCHNICK = 401
ERR_NOSUCHSERVER = 402
ERR_NOSUCHCHANNEL = 403
ERR_CANNOTSENDTOCHAN = 404
ERR_TOOMANYCHANNELS = 405
ERR_WASNOSUCHNICK = 406
ERR_TOOMANYTARGETS = 407
ERR_NOORIGIN = 409
ERR_NORECIPIENT = 411
ERR_NOTEXTTOSEND = 412
ERR_NOTOPLEVEL = 413
ERR_WILDTOPLEVEL = 414
ERR_UNKNOWNCOMMAND = 421
ERR_NOMOTD = 422
ERR_NOADMININFO = 423
ERR_FILEERROR = 424
ERR_NONICKNAMEGIVEN = 431
ERR_ERRONEUSNICKNAME = 432
ERR_NICKNAMEINUSE = 433
ERR_NICKCOLLISION = 436
ERR_NOTONCHANNEL = 442
ERR_USERONCHANNEL = 443
ERR_NOLOGIN = 444
ERR_SUMMONDISABLED = 445
ERR_USERSDISABLED = 446
ERR_NOTREGISTERED = 451
ERR_NEEDMOREPARAMS = 461
ERR_ALREADYREGISTRED = 462
ERR_PASSWDMISMATCH = 464
ERR_YOUREBANNEDCREEP = 465
ERR_KEYSET = 467
ERR_CHANNELISFULL = 471
ERR_UNKNOWNMODE = 472
ERR_INVITEONLYCHAN = 473
ERR_BANNEDFROMCHAN = 474
ERR_BADCHANNELKEY = 475
ERR_NOPRIVILEGES = 481
ERR_CHANOPRIVSNEEDED = 482
ERR_CANTKILLSERVER = 483
ERR_NOOPERHOST = 491

ERR_UMODEUNKNOWNFLAG = 501
ERR_USERSDONTMATCH = 502



"""
STOMP Protocol

This package contains a component implementing the STOMP Client protocol.
This can be used with products such as ActiveMQ, RabbitMQ, etc
"""

from .client import StompClient


__all__ = ('StompClient',)
# pylama:skip=1



"""Development Task"""

from os import getcwd

import docker  # noqa
import help  # noqa
from fabric.api import abort, cd, execute, hide, hosts, local, prefix, prompt, run, settings, task

import docs  # noqa

from .utils import msg, pip, requires, tobool


@task()
@requires('pip')
def build(**options):
    """
    Build and install required dependencies

    Options can be provided to customize the build.
    The following options are supported:

    - dev -> Whether to install in development mode (Default: Fase)
    """
    dev = tobool(options.get('dev', False))

    if dev:
        pip(requirements='requirements-dev.txt')

    with settings(hide('stdout', 'stderr'), warn_only=True):
        local('python setup.py {:s}'.format('develop' if dev else 'install'))


@task()
def clean():
    """Clean up build files and directories"""
    files = ['build', '.coverage', 'coverage', 'dist', 'docs/build']

    local('rm -rf {:s}'.format(' '.join(files)))

    local("find . -type f -name '*~' -delete")
    local("find . -type f -name '*.pyo' -delete")
    local("find . -type f -name '*.pyc' -delete")
    local("find . -type d -name '__pycache__' -delete")
    local("find . -type d -name '*egg-info' -exec rm -rf {} +")


@task()
def develop():
    """Build and Install in Development Mode"""
    return execute(build, dev=True)


@task()
@requires('pytest')
def test():
    """Run all unit tests and doctests."""
    local('python setup.py test')


@task()
@hosts('localhost')
def release():
    """Performs a full release"""
    with cd(getcwd()):
        with msg('Creating env'):
            run('mkvirtualenv test')

        with msg('Building'), prefix('workon test'):
            run('fab develop')

        with msg('Running tests'), prefix('workon test'):
            run('fab test')

        with msg('Building docs'), prefix('workon test'):
            run('pip install -r docs/requirements.txt')
            run('fab docs')

        version = run('python setup.py --version')
        if 'dev' in version:
            abort('Detected Development Version!')

        print(f'Release version: {version:s}')

        if prompt('Is this ok?', default='Y', validate=r'^[YyNn]?$') in 'yY':
            run(f'git tag {version:s}')
            run('python setup.py egg_info sdist bdist_egg bdist_wheel register upload')
            run('python setup.py build_sphinx upload_sphinx')

        with msg('Destroying env'):
            run('rmvirtualenv test')



#!/usr/bin/env python
from circuits import Component
from circuits.web import XMLRPC, Logger, Server


class Test(Component):
    def foo(self, a, b, c):
        return a, b, c


app = Server(('0.0.0.0', 8000))
Logger().register(app)
XMLRPC().register(app)
Test().register(app)
app.run()



from grp import getgrnam
from os import getuid, setgid, setgroups, setuid, umask
from pwd import getpwnam
from traceback import format_exc

from circuits.core import BaseComponent, handler


class DropPrivileges(BaseComponent):
    def init(self, user='nobody', group='nobody', umask=0o077, **kwargs):
        self.user = user
        self.group = group
        self.umask = umask

    def drop_privileges(self):
        if getuid() > 0:
            # Running as non-root. Ignore.
            return

        try:
            # Get the uid/gid from the name
            uid = getpwnam(self.user).pw_uid
            gid = getgrnam(self.group).gr_gid
        except KeyError as error:
            print(f'ERROR: Could not drop privileges {error}')
            print(format_exc())
            raise SystemExit(-1)

        try:
            # Remove group privileges
            setgroups([])

            # Try setting the new uid/gid
            setgid(gid)
            setuid(uid)

            if self.umask is not None:
                umask(self.umask)
        except Exception as error:
            print(f'ERROR: Could not drop privileges {error}')
            print(format_exc())
            raise SystemExit(-1)

    @handler('ready', channel='*')
    def on_ready(self, server, bind):
        try:
            self.drop_privileges()
        finally:
            self.unregister()



"""Circuits component for handling Stomp Connection"""

import logging
import ssl
import time
import traceback

from circuits import BaseComponent, Timer
from circuits.core.handlers import handler
from circuits.protocols.stomp.events import (
    client_heartbeat,
    connected,
    connection_failed,
    disconnected,
    heartbeat_timeout,
    message,
    on_stomp_error,
    server_heartbeat,
)
from circuits.protocols.stomp.transport import EnhancedStompFrameTransport


try:
    from stompest.config import StompConfig
    from stompest.error import StompConnectionError, StompError
    from stompest.protocol import StompSession, StompSpec
    from stompest.sync import Stomp
    from stompest.sync.client import LOG_CATEGORY
except ImportError:
    raise ImportError('No stomp support available.  Is stompest installed?')


StompSpec.DEFAULT_VERSION = '1.2'
ACK_CLIENT_INDIVIDUAL = StompSpec.ACK_CLIENT_INDIVIDUAL
ACK_AUTO = StompSpec.ACK_AUTO
ACK_CLIENT = StompSpec.ACK_CLIENT
ACK_MODES = (ACK_CLIENT_INDIVIDUAL, ACK_AUTO, ACK_CLIENT)

LOG = logging.getLogger(__name__)


class StompClient(BaseComponent):
    """Send and Receive messages from a STOMP queue"""

    channel = 'stomp'

    def init(
        self,
        host,
        port,
        username=None,
        password=None,
        connect_timeout=3,
        connected_timeout=3,
        version=StompSpec.VERSION_1_2,
        accept_versions=['1.0', '1.1', '1.2'],
        heartbeats=(0, 0),
        ssl_context=None,
        use_ssl=True,
        key_file=None,
        cert_file=None,
        ca_certs=None,
        ssl_version=ssl.PROTOCOL_TLS,
        key_file_password=None,
        proxy_host=None,
        proxy_port=None,
        proxy_user=None,
        proxy_password=None,
        channel=channel,
    ):
        """Initialize StompClient.  Called after __init__"""
        self.channel = channel
        if proxy_host:
            LOG.info('Connect to %s:%s through proxy %s:%d', host, port, proxy_host, proxy_port)
        else:
            LOG.info('Connect to %s:%s', host, port)

        if use_ssl and not ssl_context:
            ssl_params = {
                'key_file': key_file,
                'cert_file': cert_file,
                'ca_certs': ca_certs,
                'ssl_version': ssl_version,
                'password': key_file_password,
            }
            LOG.info('Request to use old-style socket wrapper: %s', ssl_params)
            ssl_context = ssl_params

        uri = f'ssl://{host}:{port}' if use_ssl else f'tcp://{host}:{port}'

        # Configure failover options so it only tries to connect once
        self._stomp_server = 'failover:(%s)?maxReconnectAttempts=1,startupMaxReconnectAttempts=1' % uri

        self._stomp_config = StompConfig(
            uri=self._stomp_server, sslContext=ssl_context, version=version, login=username, passcode=password
        )

        self._heartbeats = heartbeats
        self._accept_versions = accept_versions
        self._connect_timeout = connect_timeout
        self._connected_timeout = connected_timeout
        Stomp._transportFactory = EnhancedStompFrameTransport
        Stomp._transportFactory.proxy_host = proxy_host
        Stomp._transportFactory.proxy_port = proxy_port
        Stomp._transportFactory.proxy_user = proxy_user
        Stomp._transportFactory.proxy_password = proxy_password
        self._client = Stomp(self._stomp_config)
        self._subscribed = {}
        self.server_heartbeat = None
        self.client_heartbeat = None
        self.ALLOWANCE = 2  # multiplier for heartbeat timeouts

    @property
    def connected(self):
        if self._client.session:
            return self._client.session.state == StompSession.CONNECTED
        return False

    @property
    def subscribed(self):
        return self._subscribed.keys()

    @property
    def stomp_logger(self):
        return LOG_CATEGORY

    @handler('disconnect')
    def _disconnect(self, receipt=None):
        if self.connected:
            self._client.disconnect(receipt=receipt)
        self._client.close(flush=True)
        self.fire(disconnected(reconnect=False))
        self._subscribed = {}
        return 'disconnected'

    def start_heartbeats(self):
        LOG.info('Client HB: %s  Server HB: %s', self._client.clientHeartBeat, self._client.serverHeartBeat)
        if self._client.clientHeartBeat:
            if self.client_heartbeat:
                # Timer already exists, just reset it
                self.client_heartbeat.reset()
            else:
                LOG.info('Client will send heartbeats to server')
                # Send heartbeats at 80% of agreed rate
                self.client_heartbeat = Timer((self._client.clientHeartBeat / 1000.0) * 0.8, client_heartbeat(), persist=True)
                self.client_heartbeat.register(self)
        else:
            LOG.info('No Client heartbeats will be sent')

        if self._client.serverHeartBeat:
            if self.server_heartbeat:
                # Timer already exists, just reset it
                self.server_heartbeat.reset()
            else:
                LOG.info('Requested heartbeats from server.')
                # Allow a grace period on server heartbeats
                self.server_heartbeat = Timer(
                    (self._client.serverHeartBeat / 1000.0) * self.ALLOWANCE, server_heartbeat(), persist=True
                )
                self.server_heartbeat.register(self)
        else:
            LOG.info('Expecting no heartbeats from Server')

    @handler('connect')
    def connect(self, event, host=None, *args, **kwargs):
        """Connect to Stomp server"""
        LOG.info('Connect to Stomp...')
        try:
            self._client.connect(
                heartBeats=self._heartbeats,
                host=host,
                versions=self._accept_versions,
                connectTimeout=self._connect_timeout,
                connectedTimeout=self._connected_timeout,
            )
            LOG.info('State after Connection Attempt: %s', self._client.session.state)
            if self.connected:
                LOG.info('Connected to %s', self._stomp_server)
                self.fire(connected())
                self.start_heartbeats()
                return 'success'

        except StompConnectionError:
            LOG.debug(traceback.format_exc())
            self.fire(connection_failed(self._stomp_server))
            event.success = False
        return 'fail'

    @handler('server_heartbeat')
    def check_server_heartbeat(self, event):
        """Confirm that heartbeat from server hasn't timed out"""
        now = time.time()
        last = self._client.lastReceived or 0
        elapsed = now - last if last else -1
        LOG.debug('Last received data %d seconds ago', elapsed)
        if ((self._client.serverHeartBeat / 1000.0) * self.ALLOWANCE + last) < now:
            LOG.error('Server heartbeat timeout. %d seconds since last heartbeat.  Disconnecting.', elapsed)
            event.success = False
            self.fire(heartbeat_timeout())
            if self.connected:
                self._client.disconnect()
            # TODO: Try to auto-reconnect?

    @handler('client_heartbeat')
    def send_heartbeat(self, event):
        if self.connected:
            LOG.debug('Sending heartbeat')
            try:
                self._client.beat()
            except StompConnectionError:
                event.success = False
                self.fire(disconnected())

    @handler('generate_events')
    def generate_events(self, event):
        if not self.connected:
            return
        try:
            if self._client.canRead(1):
                frame = self._client.receiveFrame()
                LOG.debug('Recieved frame %s', frame)
                self.fire(message(frame))
        except StompConnectionError:
            self.fire(disconnected())

    @handler('send')
    def send(self, event, destination, body, headers=None, receipt=None):
        LOG.debug('send()')
        if not self.connected:
            LOG.error("Can't send when Stomp is disconnected")
            self.fire(on_stomp_error(None, Exception('Message send attempted with stomp disconnected')))
            event.success = False
            return
        try:
            self._client.send(destination, body=body.encode('utf-8'), headers=headers, receipt=receipt)
            LOG.debug('Message sent')
        except StompConnectionError:
            event.success = False
            self.fire(disconnected())
        except StompError as err:
            LOG.error('Error sending ack')
            event.success = False
            self.fire(on_stomp_error(None, err))

    @handler('subscribe')
    def _subscribe(self, event, destination, ack=ACK_CLIENT_INDIVIDUAL):
        if ack not in ACK_MODES:
            raise ValueError('Invalid client ack mode specified')
        LOG.info('Subscribe to message destination %s', destination)
        try:
            # Set ID to match destination name for easy reference later
            _frame, token = self._client.subscribe(destination, headers={StompSpec.ACK_HEADER: ack, 'id': destination})
            self._subscribed[destination] = token
        except StompConnectionError:
            event.success = False
            self.fire(disconnected())
        except StompError as err:
            event.success = False
            LOG.debug(traceback.format_exc())
            self.fire(on_stomp_error(None, err))

    @handler('unsubscribe')
    def _unsubscribe(self, event, destination):
        if destination not in self._subscribed:
            LOG.error('Unsubscribe Request Ignored. Not subscribed to %s', destination)
            return
        try:
            token = self._subscribed.pop(destination)
            frame = self._client.unsubscribe(token)
            LOG.debug('Unsubscribed: %s', frame)
        except StompConnectionError:
            event.success = False
            self.fire(disconnected())
        except StompError as err:
            LOG.error('Error sending ack')
            event.success = False
            self.fire(on_stomp_error(frame, err))

    @handler('message')
    def on_message(self, event, headers, message):
        LOG.info('Stomp message received')

    @handler('ack')
    def ack_frame(self, event, frame):
        LOG.debug('ack_frame()')
        try:
            self._client.ack(frame)
            LOG.debug('Ack Sent')
        except StompConnectionError:
            LOG.error('Error sending ack')
            event.success = False
            self.fire(disconnected())
        except StompError as err:
            LOG.error('Error sending ack')
            event.success = False
            self.fire(on_stomp_error(frame, err))

    def get_subscription(self, frame):
        """Get subscription from frame"""
        LOG.info(self._subscribed)
        _, token = self._client.message(frame)
        return self._subscribed[token]



#!/usr/bin/env python

import sys

from circuits import Component, handler
from circuits.io import File
from circuits.net.events import connect, write
from circuits.net.sockets import TCPClient


class Telnet(Component):
    channel = 'telnet'

    def init(self, host, port):
        self.host = host
        self.port = port

        TCPClient(channel=self.channel).register(self)
        File(sys.stdin, channel='stdin').register(self)

    def ready(self, socket):
        self.fire(connect(self.host, self.port))

    def read(self, data):
        print(data.strip())

    @handler('read', channel='stdin')
    def read_user_input(self, data):
        self.fire(write(data))


host = sys.argv[1]
port = int(sys.argv[2])

Telnet(host, port).run()



import pytest

from circuits.net.sockets import TCPServer


try:
    from socket import SO_REUSEPORT, SOL_SOCKET
except ImportError:
    pytestmark = pytest.mark.skip(reason='Missing SO_REUSEPORT')


def test_socket_options_server():
    s = TCPServer(('0.0.0.0', 8090), socket_options=[(SOL_SOCKET, SO_REUSEPORT, 1)])
    assert s._sock.getsockopt(SOL_SOCKET, SO_REUSEPORT) == 1



#!/usr/bin/env python
"""
Simple Timers

A trivial simple example of using circuits and timers.
"""

from circuits import Component, Event, Timer


class App(Component):
    def hello(self):
        """
        hello Event handler

        Fired once in 5 seconds.
        """
        print('Hello World')

    def foo(self):
        """
        foo Event handler

        Fired every 1 seconds.
        """
        print('Foo')

    def bar(self):
        """
        bar Event handler

        Fired every 3 seconds.
        """
        print('Bar')

    def started(self, component):
        """
        started Event handler

        Setup 3 timers at 5, 1 and 3 seconds.
        The 2nd two timers are persitent meaning that
        they are fired repeatedly every 1 and 3 seconds
        respectively.
        """
        # Timer(seconds, event, persist=False)
        Timer(5, Event.create('hello')).register(self)
        Timer(1, Event.create('foo'), persist=True).register(self)
        Timer(3, Event.create('bar'), persist=True).register(self)


App().run()



from circuits import Component
from circuits.net.events import write


class Server(Component):
    channel = 'server'

    def init(self):
        self.data = ''
        self.host = None
        self.port = None
        self.client = None
        self.ready = False
        self.closed = False
        self.connected = False
        self.disconnected = False

    def ready(self, server, bind):
        self.ready = True
        self.host, self.port = bind

    def close(self):
        return

    def closed(self):
        self.closed = True

    def connect(self, sock, *args):
        self.connected = True
        self.client = args
        self.fire(write(sock, b'Ready'))

    def disconnect(self, sock):
        self.client = None
        self.disconnected = True

    def read(self, sock, data):
        self.data = data
        return data



#!/usr/bin/env python
import select
import socket

import pytest

from circuits import Manager
from circuits.core.pollers import EPoll, KQueue, Poll, Select
from circuits.net.events import close, write
from circuits.net.sockets import UDP6Client, UDP6Server, UDPClient, UDPServer

from .client import Client
from .server import Server


def wait_host(server):
    def checker(obj, attr):
        return all(getattr(obj, a) for a in attr)

    assert pytest.wait_for(server, ('host', 'port'), checker)


def pytest_generate_tests(metafunc):
    ipv6 = [False]
    if socket.has_ipv6:
        ipv6.append(True)

    for ipv6 in ipv6:
        poller = [(Select, ipv6)]

        if hasattr(select, 'poll'):
            poller.append((Poll, ipv6))

        if hasattr(select, 'epoll'):
            poller.append((EPoll, ipv6))

        if hasattr(select, 'kqueue'):
            poller.append((KQueue, ipv6))

    metafunc.parametrize('Poller,ipv6', poller)


def test_basic(Poller, ipv6):
    m = Manager() + Poller()

    if ipv6:
        udp_server = UDP6Server(('::1', 0))
        udp_client = UDP6Client(('::1', 0), channel='client')
    else:
        udp_server = UDPServer(0)
        udp_client = UDPClient(0, channel='client')
    server = Server() + udp_server
    client = Client() + udp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(server, 'ready')
        assert pytest.wait_for(client, 'ready')
        wait_host(server)

        client.fire(write((server.host, server.port), b'foo'))
        assert pytest.wait_for(server, 'data', b'foo')

        client.fire(close())
        assert pytest.wait_for(client, 'closed')

        server.fire(close())
        assert pytest.wait_for(server, 'closed')
    finally:
        m.stop()


def test_close(Poller, ipv6):
    m = Manager() + Poller()
    server = Server() + UDPServer(0)
    server.register(m)
    m.start()

    try:
        assert pytest.wait_for(server, 'ready')
        wait_host(server)

        host, port = server.host, server.port

        server.fire(close())
        assert pytest.wait_for(server, 'disconnected')

        server.unregister()

        def test(obj, attr):
            return attr not in obj.components

        assert pytest.wait_for(m, server, value=test)

        server = Server() + UDPServer((host, port))
        server.register(m)

        assert pytest.wait_for(server, 'ready', timeout=30.0)
    finally:
        m.stop()



#!/usr/bin/python -i
from circuits import Component, Event, Manager, handler


class test(Event):
    """test Event"""


class App(Component):
    @handler('test')
    def test_0(self):
        return 0

    @handler('test', priority=3)
    def test_3(self):
        return 3

    @handler('test', priority=2)
    def test_2(self):
        return 2


m = Manager()
app = App()
app.register(m)

while len(m):
    m.flush()


def test_main():
    v = m.fire(test())
    while len(m):
        m.flush()
    x = list(v)
    assert x == [3, 2, 0]



"""Application Components

Contains various components useful for application development and tasks
common to applications.
"""

from .daemon import Daemon
from .dropprivileges import DropPrivileges


__all__ = ('Daemon', 'DropPrivileges')

# flake8: noqa
# pylama: skip=1



"""
WSGI Components

This module implements WSGI Components.
"""

from io import StringIO
from operator import itemgetter
from sys import exc_info as _exc_info
from traceback import format_tb
from types import GeneratorType

from circuits.core import BaseComponent, handler
from circuits.web import wrappers

from .dispatchers import Dispatcher
from .errors import httperror
from .events import request
from .headers import Headers
from .http import HTTP


def create_environ(errors, path, req):
    environ = {}
    env = environ.__setitem__

    env('REQUEST_METHOD', req.method)
    env('SERVER_NAME', req.host.split(':', 1)[0])
    env('SERVER_PORT', '%i' % (req.server.port or 0))
    env('SERVER_PROTOCOL', 'HTTP/%d.%d' % req.server.http.protocol)
    env('QUERY_STRING', req.qs)
    env('SCRIPT_NAME', req.script_name)
    env('CONTENT_TYPE', req.headers.get('Content-Type', ''))
    env('CONTENT_LENGTH', req.headers.get('Content-Length', ''))
    env('REMOTE_ADDR', req.remote.ip)
    env('REMOTE_PORT', '%i' % (req.remote.port or 0))
    env('wsgi.version', (1, 0))
    env('wsgi.input', req.body)
    env('wsgi.errors', errors)
    env('wsgi.multithread', False)
    env('wsgi.multiprocess', False)
    env('wsgi.run_once', False)
    env('wsgi.url_scheme', req.scheme)

    if req.path:
        req.script_name = req.path[: len(path)]
        req.path = req.path[len(path) :]
        env('SCRIPT_NAME', req.script_name)
        env('PATH_INFO', req.path)

    for k, v in list(req.headers.items()):
        env('HTTP_%s' % k.upper().replace('-', '_'), v)

    return environ


class Application(BaseComponent):
    channel = 'web'

    headerNames = {
        'HTTP_CGI_AUTHORIZATION': 'Authorization',
        'CONTENT_LENGTH': 'Content-Length',
        'CONTENT_TYPE': 'Content-Type',
        'REMOTE_HOST': 'Remote-Host',
        'REMOTE_ADDR': 'Remote-Addr',
    }

    def init(self):
        self._finished = False

        HTTP(self).register(self)
        Dispatcher().register(self)

    def translateHeaders(self, environ):
        for cgiName in environ:
            # We assume all incoming header keys are uppercase already.
            if cgiName in self.headerNames:
                yield self.headerNames[cgiName], environ[cgiName]
            elif cgiName[:5] == 'HTTP_':
                # Hackish attempt at recovering original header names.
                translatedHeader = cgiName[5:].replace('_', '-')
                yield translatedHeader, environ[cgiName]

    def getRequestResponse(self, environ):
        env = environ.get

        headers = Headers(list(self.translateHeaders(environ)))

        protocol = tuple(map(int, env('SERVER_PROTOCOL')[5:].split('.')))
        req = wrappers.Request(
            None,
            env('REQUEST_METHOD'),
            env('wsgi.url_scheme'),
            env('PATH_INFO', ''),
            protocol,
            env('QUERY_STRING', ''),
            headers=headers,
        )

        req.remote = wrappers.Host(env('REMOTE_ADDR'), env('REMTOE_PORT'))
        req.script_name = env('SCRIPT_NAME')
        req.wsgi_environ = environ

        try:
            cl = int(headers.get('Content-Length', '0'))
        except ValueError:
            cl = 0

        req.body.write(env('wsgi.input').read(cl))  # FIXME: what about chunked encoding?
        req.body.seek(0)

        res = wrappers.Response(req)
        res.gzip = 'gzip' in req.headers.get('Accept-Encoding', '')

        return req, res

    def __call__(self, environ, start_response, exc_info=None):
        self.request, self.response = self.getRequestResponse(environ)
        self.fire(request(self.request, self.response))

        self._finished = False
        while self._queue or not self._finished:
            self.tick()

        self.response.prepare()
        body = self.response.body
        status = self.response.status
        headers = list(self.response.headers.items())

        start_response(str(status), headers, exc_info)
        return body

    @handler('response', channel='web')
    def on_response(self, event, response):
        self._finished = True
        event.stop()

    @property
    def host(self):
        return ''

    @property
    def port(self):
        return 0

    @property
    def secure(self):
        return False


class _Empty(str):
    __slots__ = ()

    def __bool__(self):
        return True

    __nonzero__ = __bool__


empty = _Empty()
del _Empty


class Gateway(BaseComponent):
    channel = 'web'

    def init(self, apps):
        self.apps = apps

        self.errors = {k: StringIO() for k in self.apps}

    @handler('request', priority=0.2)
    def _on_request(self, event, req, res):
        if not self.apps:
            return None

        parts = req.path.split('/')

        candidates = []
        for i in range(len(parts)):
            k = '/'.join(parts[: (i + 1)]) or '/'
            if k in self.apps:
                candidates.append((k, self.apps[k]))
        candidates = sorted(candidates, key=itemgetter(0), reverse=True)

        if not candidates:
            return None

        path, app = candidates[0]

        buffer = StringIO()

        def start_response(status, headers, exc_info=None):
            res.status = int(status.split(' ', 1)[0])
            for header in headers:
                res.headers.add_header(*header)
            return buffer.write

        errors = self.errors[path]

        environ = create_environ(errors, path, req)

        try:
            body = app(environ, start_response)
            if isinstance(body, list):
                _body = type(body[0])() if body else ''
                body = _body.join(body)
            elif isinstance(body, GeneratorType):
                res.body = body
                res.stream = True
                return res

            if not body:
                if not buffer.tell():
                    return empty
                buffer.seek(0)
                return buffer
            return body
        except Exception:
            etype, evalue, etraceback = _exc_info()
            error = (etype, evalue, format_tb(etraceback))
            return httperror(req, res, 500, error=error)
        finally:
            event.stop()



from circuits.web import Controller, expose

from .helpers import urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'

    @expose('+test')
    def test(self):
        return 'test'

    @expose('foo+bar', 'foo_bar')
    def foobar(self):
        return 'foobar'


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    f = urlopen('%s/+test' % webapp.server.http.base)
    s = f.read()
    assert s == b'test'

    f = urlopen('%s/foo+bar' % webapp.server.http.base)
    s = f.read()
    assert s == b'foobar'

    f = urlopen('%s/foo_bar' % webapp.server.http.base)
    s = f.read()
    assert s == b'foobar'



"""
Node

this module manage node (start server, add peer, ...)
.. seealso:: Examples in :file:`examples/node`
"""

from circuits import BaseComponent, Timer, handler
from circuits.net.events import connect

from .client import Client
from .events import connected_to, disconnected_from, remote
from .server import Server


class Node(BaseComponent):
    """
    Node

    this class manage node (start server, add peer, ...)
    .. seealso:: Examples in :file:`examples/node`
    """

    channel = 'node'
    __peers = {}

    def __init__(self, port=None, channel=channel, **kwargs):
        """
        Start node system.

        :param port:    An optional keyword argument which if defined,
                        start server on this port.
                        **Default:** ``None`` (don't start the server)
        :type port:     int

        :param server_ip:   An optional keyword argument which define
                            ip where the socket has listen to.
                            **Default:** ``0.0.0.0`` (all ip is allowed)
        :type server_ip:     str

        :param channel: An optional keyword argument which if defined,
                        set channel used for node event. **Default:** ``node``
        :type channel:  str

        :param receive_event_firewall:  An optional keyword argument which if
                                        defined, set function or method to call
                                        to check if event is allowed for sending.
                                        **Default:** ``None`` (no firewall)
        :type receive_event_firewall:   function
        :type receive_event_firewall:   method

        :param send_event_firewall:  An optional keyword argument which if
                                    defined, set function or method to call to
                                    check if event is allowed for executing
                                    **Default:** ``None`` (no firewall)
        :type send_event_firewall:   function
        :type send_event_firewall:   method
        """
        super().__init__(channel=channel, **kwargs)

        if port is not None:
            self.server = Server(port, channel=channel, **kwargs).register(self)
        else:
            self.server = None

    def add(self, connection_name, hostname, port, **kwargs):
        """
        Add new peer to the node.

        :param connection_name:    Connection name.
        :type connection_name:     str

        :param hostname:    hostname of the remote node.
        :type hostname:     str

        :param port:    port of the remote node.
        :type port:     int

        :param auto_remote_event:   An optional keyword argument which if
                                    defined, bind events automatically to remote
                                    execution. **Default:** ``{}`` (no events)
        :type auto_remote_event:    dict

        :param channel: An optional keyword argument which if defined,
                        set channel used for client event. If this keyword is
                        not defined the method will generate the channel name
                        automatically.
        :type channel:  str

        :param reconnect_delay: An optional keyword argument which if defined,
                                set auto reconnect delay.
                                **Default:** ``10`` (seconde)
        :type reconnect_delay:  int

        :param receive_event_firewall:  An optional keyword argument which if
                                        defined, function or method to call for
                                        check if event is allowed for sending.
                                        **Default:** ``None`` (no firewall)
        :type receive_event_firewall:   method

        :param send_event_firewall: An optional keyword argument which if
                                    defined, setfunction or method to call to
                                    check if event is allowed for executing
                                    **Default:** ``None`` (no firewall)
        :type send_event_firewall:  method

        :return: Channel used on client event.
        :rtype: str
        """
        # automatic send event to peer
        auto_remote_event = kwargs.pop('auto_remote_event', {})
        for event_name in auto_remote_event:
            for channel in auto_remote_event[event_name]:

                @handler(event_name, channel=channel)
                def event_handle(self, event, *args, **kwargs):
                    yield self.call(remote(event, connection_name))

            self.addHandler(event_handle)

        client_channel = kwargs.pop(
            'channel',
            f'{self.channel}_client_{connection_name}',
        )
        reconnect_delay = kwargs.pop('reconnect_delay', 10)
        client = Client(hostname, port, channel=client_channel, **kwargs)

        # connected event binding
        @handler('connected', channel=client_channel)
        def connected(self, hostname, port):
            self.fire(
                connected_to(
                    connection_name,
                    hostname,
                    port,
                    client_channel,
                    client,
                )
            )

        self.addHandler(connected)

        # disconnected event binding
        @handler('disconnected', 'unreachable', channel=client_channel)
        def disconnected(self, event, *args, **kwargs):
            if event.name == 'disconnected':
                self.fire(
                    disconnected_from(
                        connection_name,
                        hostname,
                        port,
                        client_channel,
                        client,
                    )
                )

            # auto reconnect
            if reconnect_delay > 0:
                Timer(
                    reconnect_delay,
                    connect(hostname, port),
                    client_channel,
                ).register(self)

        self.addHandler(disconnected)

        client.register(self)
        self.__peers[connection_name] = client
        return client_channel

    def get_connection_names(self):
        """
        Get connections names

        :return: The list of connections names
        :rtype: list of str
        """
        return list(self.__peers)

    def get_peer(self, connection_name):
        """
        Get a client object by name

        :param connection_name:    Connection name.
        :type connection_name:     str

        :return: The Client object
        :rtype: :class:`circuits.node.client.Client`
        """
        return self.__peers.get(connection_name, None)

    @handler('remote', channel='*')
    def __on_remote(self, event, remote_event, connection_name, channel=None):
        """
        Send event to peer

        Event handler to run an event on peer (the event definition is
        :class:`circuits.node.events.remote`)

        :param event:    The event triggered (by the handler)
        :type event:     :class:`circuits.node.events.remote`

        :param remote_event:    Event to execute remotely.
        :type remote_event:     :class:`circuits.core.events.Event`

        :param connection_name:    Connection name (peer selection).
        :type connection_name:     str

        :param channel:    Remote channel (channel to use on peer).
        :type channel:     str

        :return: The result of remote event
        :rtype: generator

        :Example:
        ``# hello is your event to execute remotely
        # peer_test is peer name
        result = yield self.fire(remote(hello())), 'peer_test')
        print(result.value)``
        """
        node = self.__peers[connection_name]
        remote_event.channels = (channel,) if channel is not None else event.channels
        return node.send(remote_event)



#!/usr/bin/env python
import pytest

from circuits.core import Component, Event, TimeoutError, handler


class wait(Event):
    """wait Event"""

    success = True


class call(Event):
    """call Event"""

    success = True


class hello(Event):
    """hello Event"""

    success = True


class App(Component):
    @handler('wait')
    def _on_wait(self, timeout=-1):
        result = self.fire(hello())
        try:
            yield self.wait('hello', timeout=timeout)
        except TimeoutError as e:
            yield e
        else:
            yield result

    @handler('hello')
    def _on_hello(self):
        return 'hello'

    @handler('call')
    def _on_call(self, timeout=-1):
        result = None
        try:
            result = yield self.call(hello(), timeout=timeout)
        except TimeoutError as e:
            yield e
        else:
            yield result


@pytest.fixture()
def app(request, manager, watcher):
    app = App().register(manager)
    assert watcher.wait('registered')

    def finalizer():
        app.unregister()

    request.addfinalizer(finalizer)

    return app


def test_wait_success(manager, watcher, app):
    x = manager.fire(wait(10))
    assert watcher.wait('wait_success')

    value = x.value

    assert value == 'hello'


def test_wait_failure(manager, watcher, app):
    x = manager.fire(wait(0))
    assert watcher.wait('wait_success')

    value = x.value

    assert isinstance(value, TimeoutError)


def test_call_success(manager, watcher, app):
    x = manager.fire(call(10))
    assert watcher.wait('call_success')

    value = x.value

    assert value == 'hello'


def test_call_failure(manager, watcher, app):
    x = manager.fire(call(0))
    assert watcher.wait('call_success')

    value = x.value

    assert isinstance(value, TimeoutError)



#!/usr/bin/env python
from glob import glob

from setuptools import find_packages, setup


def read_file(filename):
    try:
        return open(filename).read()
    except OSError:
        return ''


setup(
    name='circuits',
    description='Asynchronous Component based Event Application Framework',
    long_description=open('README.rst')
    .read()
    .replace(
        '.. include:: examples/index.rst',
        read_file('examples/index.rst'),
    ),
    author='James Mills',
    author_email='prologic@shortcircuit.net.au',
    url='http://circuitsframework.com/',
    download_url='http://bitbucket.org/circuits/circuits/downloads/',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Environment :: Other Environment',
        'Environment :: Plugins',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Adaptive Technologies',
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Communications :: Email :: Mail Transport Agents',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
    ],
    license='MIT',
    keywords='event framework distributed concurrent component asynchronous',
    platforms='POSIX',
    packages=find_packages(
        exclude=[
            '*.tests',
            '*.tests.*',
            'tests.*',
            'tests',
            '*.fabfile',
            '*.fabfile.*',
            'fabfile.*',
            'fabfile',
        ],
    ),
    scripts=glob('bin/*'),
    entry_points={
        'console_scripts': [
            'circuits.web=circuits.web.main:main',
        ],
    },
    zip_safe=True,
    use_scm_version={
        'write_to': 'circuits/version.py',
    },
    setup_requires=[
        'setuptools_scm',
    ],
    extras_require={'stomp': ['stompest>=2.3.0', 'pysocks>=1.6.7']},
    python_requires='>=3.7',
)



"""Lightweight Event driven and Asynchronous Application Framework

circuits is a **Lightweight** **Event** driven and **Asynchronous**
**Application Framework** for the `Python Programming Language`_
with a strong **Component** Architecture.

:copyright: CopyRight (C) 2004-2016 by James Mills
:license: MIT (See: LICENSE)

.. _Python Programming Language: http://www.python.org/
"""

__author__ = 'James Mills'
__date__ = '24th February 2013'

try:
    from .version import version as __version__
except ImportError:
    __version__ = 'unknown'

from .core import (
    BaseComponent,
    Bridge,
    Component,
    Debugger,
    Event,
    Loader,
    Manager,
    TimeoutError,
    Timer,
    Worker,
    handler,
    ipc,
    reprhandler,
    sleep,
    task,
)


# See http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    import importlib.metadata
    import sys

    try:
        namespace_pkg = importlib.metadata.distribution(__name__)
    except importlib.metadata.PackageNotFoundError:
        pass
    else:
        namespace_pkg_files = namespace_pkg.files
        if namespace_pkg_files:
            for file in namespace_pkg_files:
                if str(file).endswith('__init__.py'):
                    namespace_pkg_path = str(file).split('__init__.py')[0]
                    if namespace_pkg_path not in sys.path:
                        sys.path.append(namespace_pkg_path)
                    break

# flake8: noqa
# pylama:skip=1
